# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Union

from bungio.models.base import (
    BaseEnum,
    BaseFlagEnum,
    BaseModel,
    ManifestModel,
    custom_define,
    custom_field,
)
from bungio.utils import enum_converter


@custom_define()
class UserTheme(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        user_theme_description: _No description given by bungie._
        user_theme_id: _No description given by bungie._
        user_theme_name: _No description given by bungie._
    """

    user_theme_description: str = custom_field()
    user_theme_id: int = custom_field()
    user_theme_name: str = custom_field()


@custom_define()
class GroupTheme(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        description: _No description given by bungie._
        folder: _No description given by bungie._
        name: _No description given by bungie._
    """

    description: str = custom_field()
    folder: str = custom_field()
    name: str = custom_field()



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import TYPE_CHECKING, Optional

from bungio.models.base import BaseModel, custom_define, custom_field

if TYPE_CHECKING:
    from bungio.models import (
        DestinyPresentationNodeDefinition,
        DestinySeasonDefinition,
        DestinyVendorDefinition,
        EmailSettings,
    )


@custom_define()
class CoreSettingsConfiguration(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        clan_banner_decal_colors: _No description given by bungie._
        clan_banner_decals: _No description given by bungie._
        clan_banner_gonfalon_colors: _No description given by bungie._
        clan_banner_gonfalon_detail_colors: _No description given by bungie._
        clan_banner_gonfalon_details: _No description given by bungie._
        clan_banner_gonfalons: _No description given by bungie._
        clan_banner_standards: _No description given by bungie._
        destiny2_core_settings: _No description given by bungie._
        destiny_membership_types: _No description given by bungie._
        email_settings: _No description given by bungie._
        environment: _No description given by bungie._
        fireteam_activities: _No description given by bungie._
        forum_categories: _No description given by bungie._
        group_avatars: _No description given by bungie._
        ignore_reasons: _No description given by bungie._
        recruitment_activities: _No description given by bungie._
        recruitment_misc_tags: _No description given by bungie._
        recruitment_platform_tags: _No description given by bungie._
        system_content_locales: _No description given by bungie._
        systems: _No description given by bungie._
        user_content_locales: _No description given by bungie._
    """

    clan_banner_decal_colors: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    clan_banner_decals: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    clan_banner_gonfalon_colors: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    clan_banner_gonfalon_detail_colors: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    clan_banner_gonfalon_details: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    clan_banner_gonfalons: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    clan_banner_standards: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    destiny2_core_settings: "Destiny2CoreSettings" = custom_field()
    destiny_membership_types: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    email_settings: "EmailSettings" = custom_field()
    environment: str = custom_field()
    fireteam_activities: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    forum_categories: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    group_avatars: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    ignore_reasons: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    recruitment_activities: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    recruitment_misc_tags: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    recruitment_platform_tags: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    system_content_locales: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    systems: dict[str, "CoreSystem"] = custom_field(metadata={"type": """dict[str, CoreSystem]"""})
    user_content_locales: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})


@custom_define()
class CoreSystem(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        enabled: _No description given by bungie._
        parameters: _No description given by bungie._
    """

    enabled: bool = custom_field()
    parameters: dict[str, str] = custom_field(metadata={"type": """dict[str, str]"""})


@custom_define()
class CoreSetting(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        child_settings: _No description given by bungie._
        display_name: _No description given by bungie._
        identifier: _No description given by bungie._
        image_path: _No description given by bungie._
        is_default: _No description given by bungie._
        summary: _No description given by bungie._
    """

    child_settings: list["CoreSetting"] = custom_field(metadata={"type": """list[CoreSetting]"""})
    display_name: str = custom_field()
    identifier: str = custom_field()
    image_path: str = custom_field()
    is_default: bool = custom_field()
    summary: str = custom_field()


@custom_define()
class Destiny2CoreSettings(BaseModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        active_seals_root_node_hash: _No description given by bungie._
        active_triumphs_root_node_hash: _No description given by bungie._
        ammo_type_heavy_icon: _No description given by bungie._
        ammo_type_primary_icon: _No description given by bungie._
        ammo_type_special_icon: _No description given by bungie._
        badges_root_node: _No description given by bungie._
        collection_root_node: _No description given by bungie._
        crafting_root_node_hash: _No description given by bungie._
        current_rank_progression_hashes: _No description given by bungie._
        current_season_hash: _No description given by bungie._
        current_seasonal_artifact_hash: _No description given by bungie._
        exotic_catalysts_root_node_hash: _No description given by bungie._
        future_season_hashes: _No description given by bungie._
        insert_plug_free_blocked_socket_type_hashes: _No description given by bungie._
        insert_plug_free_protected_plug_item_hashes: _No description given by bungie._
        legacy_seals_root_node_hash: _No description given by bungie._
        legacy_triumphs_root_node_hash: _No description given by bungie._
        lore_root_node_hash: _No description given by bungie._
        medals_root_node: _No description given by bungie._
        medals_root_node_hash: _No description given by bungie._
        metrics_root_node: _No description given by bungie._
        past_season_hashes: _No description given by bungie._
        records_root_node: _No description given by bungie._
        seasonal_challenges_presentation_node_hash: _No description given by bungie._
        undiscovered_collectible_image: _No description given by bungie._
        manifest_active_seals_root_node_hash: Manifest information for `active_seals_root_node_hash`
        manifest_active_triumphs_root_node_hash: Manifest information for `active_triumphs_root_node_hash`
        manifest_badges_root_node: Manifest information for `badges_root_node`
        manifest_collection_root_node: Manifest information for `collection_root_node`
        manifest_crafting_root_node_hash: Manifest information for `crafting_root_node_hash`
        manifest_current_season_hash: Manifest information for `current_season_hash`
        manifest_current_seasonal_artifact_hash: Manifest information for `current_seasonal_artifact_hash`
        manifest_exotic_catalysts_root_node_hash: Manifest information for `exotic_catalysts_root_node_hash`
        manifest_legacy_seals_root_node_hash: Manifest information for `legacy_seals_root_node_hash`
        manifest_legacy_triumphs_root_node_hash: Manifest information for `legacy_triumphs_root_node_hash`
        manifest_lore_root_node_hash: Manifest information for `lore_root_node_hash`
        manifest_medals_root_node: Manifest information for `medals_root_node`
        manifest_medals_root_node_hash: Manifest information for `medals_root_node_hash`
        manifest_metrics_root_node: Manifest information for `metrics_root_node`
        manifest_records_root_node: Manifest information for `records_root_node`
        manifest_seasonal_challenges_presentation_node_hash: Manifest information for `seasonal_challenges_presentation_node_hash`
    """

    active_seals_root_node_hash: int = custom_field()
    active_triumphs_root_node_hash: int = custom_field()
    ammo_type_heavy_icon: str = custom_field()
    ammo_type_primary_icon: str = custom_field()
    ammo_type_special_icon: str = custom_field()
    badges_root_node: int = custom_field()
    collection_root_node: int = custom_field()
    crafting_root_node_hash: int = custom_field()
    current_rank_progression_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    current_season_hash: int = custom_field()
    current_seasonal_artifact_hash: int = custom_field()
    exotic_catalysts_root_node_hash: int = custom_field()
    future_season_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    insert_plug_free_blocked_socket_type_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    insert_plug_free_protected_plug_item_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    legacy_seals_root_node_hash: int = custom_field()
    legacy_triumphs_root_node_hash: int = custom_field()
    lore_root_node_hash: int = custom_field()
    medals_root_node: int = custom_field()
    medals_root_node_hash: int = custom_field()
    metrics_root_node: int = custom_field()
    past_season_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    records_root_node: int = custom_field()
    seasonal_challenges_presentation_node_hash: int = custom_field()
    undiscovered_collectible_image: str = custom_field()
    manifest_active_seals_root_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_active_triumphs_root_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_badges_root_node: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_collection_root_node: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_crafting_root_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_current_season_hash: Optional["DestinySeasonDefinition"] = custom_field(default=None)
    manifest_current_seasonal_artifact_hash: Optional["DestinyVendorDefinition"] = custom_field(default=None)
    manifest_exotic_catalysts_root_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_legacy_seals_root_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_legacy_triumphs_root_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_lore_root_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_medals_root_node: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_medals_root_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_metrics_root_node: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_records_root_node: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)
    manifest_seasonal_challenges_presentation_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(
        default=None
    )



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import Optional

from bungio.models import CoreSettingsConfiguration
from bungio.models.auth import AuthData
from bungio.models.base import ClientMixin, custom_define


@custom_define()
class SettingsRouteInterface(ClientMixin):
    async def get_common_settings(self, auth: Optional[AuthData] = None) -> CoreSettingsConfiguration:
        """
        Get the common settings used by the Bungie.Net environment.

        Args:
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_common_settings(auth=auth)
        return await CoreSettingsConfiguration.from_dict(data=response, client=self._client, auth=auth)



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Optional, Union

from bungio.models.base import BaseModel, custom_define, custom_field
from bungio.utils import enum_converter

if TYPE_CHECKING:
    from bungio.models import BungieMembershipType

__all__ = ("AuthData",)


@custom_define()
class AuthData(BaseModel):
    """
    Bungie User Authentication Information

    Attributes:
        token: The token. If None, the expiry date has passed, and it has been invalidated.
        token_expiry: The token expiry date
        refresh_token: The refresh token
        refresh_token_expiry: The refresh token expiry date
        membership_type: The `membership_type` of the user this data belongs to
        membership_id: The `membership_id` of the user this data belongs to
        bungie_name: The bungie name of the user this data belongs to. Can be `None` if the user has not launched the game after they got introduced
        cross_save_setup: If the user has cross save set up. If this is `False`, the membership information may be for an unwanted system
    """

    token: Optional[str] = custom_field(repr=False)
    token_expiry: datetime.datetime = custom_field()
    refresh_token: str = custom_field(repr=False)
    refresh_token_expiry: datetime.datetime = custom_field()
    membership_type: Union["BungieMembershipType", int] = custom_field(converter=enum_converter("BungieMembershipType"))
    membership_id: int = custom_field()
    bungie_name: Optional[str] = custom_field()
    cross_save_setup: bool = custom_field(default=True)

    async def refresh(self) -> AuthData:
        """
        Check if tokens need to be refreshed and then do that.
        Gets called automatically when doing requests with AuthData.

        Tip: Staying up to date
            This dispatches the `Client.on_token_update()` event

        Args:
            auth: The potentially old authentication info.

        Raises:
            InvalidAuthentication: If authentication is invalid

        Returns:
            The working authentication info.
        """

        await self._client.get_working_auth(auth=self)



from __future__ import annotations

import builtins
import datetime
import inspect
from enum import Enum, IntFlag
from functools import partial
from typing import TYPE_CHECKING, Any, Optional, Type, _UnionGenericAlias

import attrs
from sqlalchemy.ext.asyncio import AsyncEngine

import bungio.models as models

if TYPE_CHECKING:
    from bungio.client import Client

__all__ = (
    "custom_define",
    "custom_field",
    "MISSING",
    "BaseEnum",
    "BaseFlagEnum",
    "BaseModel",
    "ManifestModel",
    "ClientMixin",
    "UnknownEnumValue",
    "FuzzyAttrFinder",
    "EnumMixin",
)


custom_define = partial(
    attrs.define,
    **{
        "slots": True,
        "kw_only": True,
    },
)
custom_field = partial(attrs.field, **{})


_catch_types = []
for entry in ["dict", "int", "str", "float", "bool"]:
    _catch_types.extend([entry, getattr(builtins, entry)])


def _enforce_type(field_type: Any, value: Any) -> Any:
    if field_type in _catch_types:
        # enforce the field_type, so the typing is correct
        # bungie does not do this
        if isinstance(field_type, str):
            field_type = getattr(builtins, field_type)
        if not isinstance(value, field_type):
            value = field_type(value)
    elif field_type in ["Any", Any]:
        pass
    else:
        # this mean it's not something we want to enforce, so raise an error
        raise ValueError
    return value


# def _transform_builtins(cls, fields: list[attrs.Attribute]) -> list[attrs.Attribute]:
#     for


class MISSING:
    """
    A missing entry which was not returned by bungie whatsoever.
    """

    def __bool__(self) -> bool:
        return False


MISSING = MISSING()


@custom_define()
class UnknownEnumValue:
    """
    Sometimes Bungie returns information that they have not defined anywhere, so this has to do
    """

    value: int | str = custom_field()
    enum: Type[BaseEnum] = custom_field()


class EnumMixin(Enum):
    """
    Mixin for enums
    """

    @staticmethod
    def process_dict(data: int | str, client: "Client", *args, **kwargs) -> int | str:
        """
        Enum specific cleanup

        Args:
            data: The int / str representation of the enum, usually received by bungie
            client: The client obj

        Returns:
            Clean int / str representation
        """
        return data

    @classmethod
    async def from_dict(cls, data: int | str, client: "Client", *args, **kwargs) -> EnumMixin | UnknownEnumValue:
        """
        Convert data to this enum

        Args:
            data: The int / str representation of the enum, usually received by bungie
            client: The client obj

        Returns:
            The enum
        """

        if isinstance(data, cls):
            return data

        data = cls.process_dict(data=data, client=client)

        # catch unknown values
        try:
            return cls(data)
        except ValueError:
            return UnknownEnumValue(value=data, enum=cls)

    def to_dict(self) -> Any:
        """
        Convert the enum into a representation bungie accepts

        Returns:
            The value which can be sent to bungie
        """

        return self.value


class BaseEnum(EnumMixin, Enum):
    """
    Base methods which help to acquire this model from json and export it to json.
    """

    @property
    def display_name(self) -> str:
        """
        Format the instance name so that it looks like in-game.

        Example:
            `name="HAND_CANNON"` -> `"Hand Cannon"`

        Returns:
            THe formatted name
        """

        return " ".join([part.capitalize() for part in self.name.split("_")])


class BaseFlagEnum(EnumMixin, IntFlag):
    """
    Base methods which help to acquire this model from json and export it to json.
    """

    pass


@custom_define()
class ClientMixin:
    """
    Mixin that give models access to the client obj
    """

    _client: "Client" = custom_field(repr=False, init=False)

    @_client.default
    def __client_factory(self) -> "Client":
        from bungio.singleton import client

        if client is MISSING:
            raise ValueError("You have to set-up your 'Client' first")
        return client


@custom_define()
class FuzzyAttrFinder:
    def _fuzzy_getattr(self, name: str) -> Any:
        """
        Returns the objs attribute that fully matches the name, or if that does not exist, the first attribute that includes the name

        Args:
            name: The name to match

        Raises:
            KeyError: If no match is found

        Returns:
            The attribute value
        """

        try:
            found_attr = getattr(self, name)
            return found_attr
        except AttributeError:
            for attr_name in self.__dir__():
                if name in attr_name:
                    return getattr(self, attr_name)
            raise KeyError(f"`{name}` not found in `{self.__dir__()}`")


@custom_define()
class BaseModel(ClientMixin):
    """
    Base methods which help to acquire this model from json and export it to json.
    """

    @staticmethod
    def process_dict(data: dict, client: "Client", *args, **kwargs) -> dict:
        """
        Model specific cleanup

        Args:
            data: The json representation of the model, usually received by bungie
            client: The client obj

        Returns:
            Clean json
        """
        return data

    @classmethod
    async def from_dict(cls, data: dict, client: "Client", recursive: bool = False, *args, **kwargs) -> BaseModel:
        """
        Convert json data to this model

        Args:
            data: The json representation of the model, usually received by bungie
            client: The client obj
            recursive: If this was called recursively

        Returns:
            The model
        """

        if isinstance(data, cls):
            return data

        if not isinstance(data, dict):
            raise ValueError

        if "Response" in data:
            data = data["Response"]

        # also use the kwargs as data
        data = kwargs | data
        data = cls.process_dict(data=data, client=client)

        prepared = {}
        for name, field in attrs.fields_dict(cls).items():
            if field.init and name != "_client":
                # get the value we want. This also skips the manifest_... entries since they have no value and a default
                value = data.get(cls._convert_to_bungie_case(name), attrs.NOTHING)
                default = field.default

                # sadly bungie sometimes does not return info without marking that fact in the api specs
                if value is attrs.NOTHING and default is attrs.NOTHING:
                    default = MISSING

                if value is attrs.NOTHING:
                    value = default

                else:
                    value = await cls._convert_to_type(
                        field_type=field.type, field_metadata=field.metadata, value=value, client=client
                    )

                # set the attr
                prepared[name] = value

        res = cls(**prepared)  # noqa

        # fill the manifest information
        if not recursive and res._client.always_return_manifest_information:
            await res.fetch_manifest_information()

        return res

    @staticmethod
    async def _convert_to_type(field_type: Any, field_metadata: Optional[dict], value: Any, client: Client) -> Any:
        # catch optional
        if isinstance(field_type, _UnionGenericAlias) and "Optional" in str(field_type):
            if value is None:
                return value
            field_type = str(field_type).removeprefix("typing.Optional[").removesuffix("]")

        # catch build-ins
        try:
            return _enforce_type(field_type=field_type, value=value)
        except ValueError:
            # sometimes the field type is the attr class as a string
            if isinstance(field_type, str):
                field_type = getattr(models, field_type)

        # convert models in models
        if hasattr(field_type, "from_dict"):
            value = await field_type.from_dict(data=value, client=client, recursive=True)

        # convert datetime
        elif field_type == datetime.datetime:
            # sometimes this includes milliseconds
            try:
                value = datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S%z")
            except ValueError:
                value = datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z")

        # convert lists / dict
        elif field_metadata:
            field_type = field_metadata["type"]

            # iterables:
            if "dict" in field_metadata["type"] or "list" in field_metadata["type"]:
                # split the type into the subtypes
                split_types = field_metadata["type"].replace("]", "").split("[")

                match split_types[0]:
                    case "dict":
                        # sometimes unknown dicts are returned
                        if field_type != "dict":
                            new_field_metadata = None
                            new_field_types = field_type.removeprefix("dict[").removesuffix("]").split(", ")
                            key_type = new_field_types[0]
                            value_type = ", ".join(new_field_types[1:])

                            # catch nested dicts
                            if "dict[" in value_type:
                                new_field_metadata = {"type": value_type}
                                value_type = None

                            ret = {}
                            for key, value in value.items():
                                key = await BaseModel._convert_to_type(
                                    field_type=key_type,
                                    field_metadata=new_field_metadata,
                                    value=key,
                                    client=client,
                                )
                                value = await BaseModel._convert_to_type(
                                    field_type=value_type,
                                    field_metadata=new_field_metadata,
                                    value=value,
                                    client=client,
                                )

                                ret[key] = value
                            value = ret

                    case "list":
                        value = [
                            await BaseModel._convert_to_type(
                                field_type=split_types[1], field_metadata=None, value=entry, client=client
                            )
                            for entry in value
                        ]

                    case _:
                        raise ValueError(f"Unexpected type {split_types[0]} in {split_types}")

            # enums
            else:
                value = await BaseModel._convert_to_type(
                    field_type=field_metadata["type"].replace('"', ""),
                    field_metadata=None,
                    value=value,
                    client=client,
                )

        return value

    @staticmethod
    def _convert_to_bungie_case(string: str) -> str:
        """
        Convert a string to how it is represented by bungie: my_name_string -> myNameString

        Args:
            string: The og string

        Returns:
            The bungie string
        """

        split = string.split("_")
        if len(split) == 1:
            return split[0]
        else:
            return "".join([split[0], *[s.capitalize() for s in split[1:]]])

    def to_dict(self) -> dict:
        """
        Convert the model into a dict representation bungie accepts

        Returns:
            A dict which can be sent to bungie
        """

        result = {}

        for name in self.__dir__():
            if name.startswith("_") or name.startswith("manifest_"):
                continue

            value = getattr(self, name)

            if inspect.ismethod(value) or inspect.isfunction(value):
                continue

            elif isinstance(value, dict):
                raise NotImplementedError(
                    "Nested dict conversion is not currently implemented, since bungie does never require that info"
                )

            elif isinstance(value, list):
                list_results = []
                for entry in value:
                    if hasattr(entry, "to_dict"):
                        list_results.append(entry.to_dict())
                    else:
                        list_results.append(entry)

                result[name] = list_results

            else:
                if hasattr(value, "to_dict"):
                    result[name] = value.to_dict()
                else:
                    result[name] = value

        return result

    async def fetch_manifest_information(self, _cache: Optional[dict] = None):
        """
        Fill the model in-place with information from the manifest.
        """

        if not isinstance(self._client.manifest_storage, AsyncEngine):
            raise ValueError("Client.manifest_storage must be set up to use this")

        if not _cache:
            _cache = {}

        class_definition = fields_dict(type(self))  # noqa

        # loop through the class attributes
        for name in self.__dir__():
            if name.startswith("__"):
                continue

            # manifest entries
            if name.startswith("manifest_"):
                striped_name = name.removeprefix("manifest_")
                value = getattr(self, striped_name)

                if value is MISSING:
                    return

                # check the cache to avoid infinite recursion
                if cached := _cache.get(value, None):
                    manifest_value = cached

                else:
                    attr_definition = class_definition[name]
                    manifest_class_name = attr_definition.type.__str__().removesuffix("')]").split("'")[-1]
                    manifest_value = await self._client.manifest.fetch(
                        manifest_class=getattr(models, manifest_class_name), value=value
                    )

                    _cache[value] = manifest_value

                    # check if the model has manifest models itself
                    if manifest_value:
                        await manifest_value.fetch_manifest_information(_cache=_cache)

                setattr(self, name, manifest_value)

            # sub models which may have manifest entries too
            elif hasattr((value := getattr(self, name)), "fetch_manifest_information"):
                await value.fetch_manifest_information(_cache=_cache)


@custom_define()
class ManifestModel(BaseModel):
    pass



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from datetime import datetime
from typing import TYPE_CHECKING, Union

from bungio.models.base import BaseModel, custom_define, custom_field
from bungio.utils import enum_converter

if TYPE_CHECKING:
    from bungio.models import DestinyItemQuantity, DestinyVendorItemRefundPolicy


@custom_define()
class DestinyVendorReceipt(BaseModel):
    """
    If a character purchased an item that is refundable, a Vendor Receipt will be created on the user's Destiny Profile. These expire after a configurable period of time, but until then can be used to get refunds on items. BNet does not provide the ability to refund a purchase *yet*, but you know.

    None
    Attributes:
        currency_paid: The amount paid for the item, in terms of items that were consumed in the purchase and their quantity.
        expires_on: The date at which this receipt is rendered invalid.
        item_received: The item that was received, and its quantity.
        license_unlock_hash: The unlock flag used to determine whether you still have the purchased item.
        purchased_by_character_id: The ID of the character who made the purchase.
        refund_policy: Whether you can get a refund, and what happens in order for the refund to be received. See the DestinyVendorItemRefundPolicy enum for details.
        sequence_number: The identifier of this receipt.
        time_to_expiration: The seconds since epoch at which this receipt is rendered invalid.
    """

    currency_paid: list["DestinyItemQuantity"] = custom_field(metadata={"type": """list[DestinyItemQuantity]"""})
    expires_on: datetime = custom_field()
    item_received: "DestinyItemQuantity" = custom_field()
    license_unlock_hash: int = custom_field()
    purchased_by_character_id: int = custom_field()
    refund_policy: Union["DestinyVendorItemRefundPolicy", int] = custom_field(
        converter=enum_converter("DestinyVendorItemRefundPolicy")
    )
    sequence_number: int = custom_field()
    time_to_expiration: int = custom_field()



from typing import Callable, Coroutine, Optional

from bungio.http.route import Route
from bungio.models.auth import AuthData


class CommunityContentRouteHttpRequests:
    request: Callable[..., Coroutine]

    async def get_community_content(
        self, media_filter: int, page: int, sort: int, auth: Optional[AuthData] = None
    ) -> dict:
        """
        Returns community content.

        Args:
            media_filter: The type of media to get
            page: Zero based page
            sort: The sort mode.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Raises:
            NotFound: 404 request
            BadRequest: 400 request
            InvalidAuthentication: If authentication is invalid
            TimeoutException: If no connection could be made
            BungieDead: Servers are down
            AuthenticationTooSlow: The authentication key has expired
            BungieException: Relaying the bungie error

        Returns:
            The json response
        """

        return await self.request(
            Route(path=f"/CommunityContent/Get/{sort}/{media_filter}/{page}/", method="GET", auth=auth)
        )



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.


from bungio.models.base import BaseModel, custom_define, custom_field


@custom_define()
class DestinyAnimationReference(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        anim_identifier: _No description given by bungie._
        anim_name: _No description given by bungie._
        path: _No description given by bungie._
    """

    anim_identifier: str = custom_field()
    anim_name: str = custom_field()
    path: str = custom_field()



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import TYPE_CHECKING

from bungio.models.base import BaseModel, custom_define, custom_field

if TYPE_CHECKING:
    pass


@custom_define()
class DestinyReportOffensePgcrRequest(BaseModel):
    """
    If you want to report a player causing trouble in a game, this request will let you report that player and the specific PGCR in which the trouble was caused, along with why. Please don't do this just because you dislike the person! I mean, I know people will do it anyways, but can you like take a good walk, or put a curse on them or something? Do me a solid and reconsider. Note that this request object doesn't have the actual PGCR ID nor your Account/Character ID in it. We will infer that information from your authentication information and the PGCR ID that you pass into the URL of the reporting endpoint itself.

    None
    Attributes:
        offending_character_id: Within the PGCR provided when calling the Reporting endpoint, this should be the character ID of the user that you thought was violating terms of use. They must exist in the PGCR provided.
        reason_category_hashes: So you've decided to report someone instead of cursing them and their descendants. Well, okay then. This is the category or categorie(s) of infractions for which you are reporting the user. These are hash identifiers that map to DestinyReportReasonCategoryDefinition entries.
        reason_hashes: If applicable, provide a more specific reason(s) within the general category of problems provided by the reasonHash. This is also an identifier for a reason. All reasonHashes provided must be children of at least one the reasonCategoryHashes provided.
    """

    offending_character_id: int = custom_field()
    reason_category_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    reason_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Union

from bungio.models.base import (
    BaseEnum,
    BaseFlagEnum,
    BaseModel,
    ManifestModel,
    custom_define,
    custom_field,
)
from bungio.models.mixins import DestinyUserMixin
from bungio.utils import enum_converter

if TYPE_CHECKING:
    from bungio.models import BungieMembershipType, GroupUserInfoCard, IgnoreResponse


@custom_define()
class UserMembership(BaseModel, DestinyUserMixin):
    """
    Very basic info about a user as returned by the Account server.

    None
    Attributes:
        bungie_global_display_name: The bungie global display name, if set.
        bungie_global_display_name_code: The bungie global display name code, if set.
        display_name: Display Name the player has chosen for themselves. The display name is optional when the data type is used as input to a platform API.
        membership_id: Membership ID as they user is known in the Accounts service
        membership_type: Type of the membership. Not necessarily the native type.
    """

    bungie_global_display_name: str = custom_field()
    bungie_global_display_name_code: int = custom_field()
    display_name: str = custom_field()
    membership_id: int = custom_field()
    membership_type: Union["BungieMembershipType", int] = custom_field(converter=enum_converter("BungieMembershipType"))


@custom_define()
class CrossSaveUserMembership(BaseModel, DestinyUserMixin):
    """
    Very basic info about a user as returned by the Account server, but including CrossSave information. Do NOT use as a request contract.

    None
    Attributes:
        applicable_membership_types: The list of Membership Types indicating the platforms on which this Membership can be used.  Not in Cross Save = its original membership type. Cross Save Primary = Any membership types it is overridding, and its original membership type Cross Save Overridden = Empty list
        bungie_global_display_name: The bungie global display name, if set.
        bungie_global_display_name_code: The bungie global display name code, if set.
        cross_save_override: If there is a cross save override in effect, this value will tell you the type that is overridding this one.
        display_name: Display Name the player has chosen for themselves. The display name is optional when the data type is used as input to a platform API.
        is_public: If True, this is a public user membership.
        membership_id: Membership ID as they user is known in the Accounts service
        membership_type: Type of the membership. Not necessarily the native type.
    """

    applicable_membership_types: list[Union["BungieMembershipType", int]] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    bungie_global_display_name: str = custom_field()
    bungie_global_display_name_code: int = custom_field()
    cross_save_override: Union["BungieMembershipType", int] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    display_name: str = custom_field()
    is_public: bool = custom_field()
    membership_id: int = custom_field()
    membership_type: Union["BungieMembershipType", int] = custom_field(converter=enum_converter("BungieMembershipType"))


@custom_define()
class UserInfoCard(BaseModel, DestinyUserMixin):
    """
    This contract supplies basic information commonly used to display a minimal amount of information about a user. Take care to not add more properties here unless the property applies in all (or at least the majority) of the situations where UserInfoCard is used. Avoid adding game specific or platform specific details here. In cases where UserInfoCard is a subset of the data needed in a contract, use UserInfoCard as a property of other contracts.

    None
    Attributes:
        applicable_membership_types: The list of Membership Types indicating the platforms on which this Membership can be used.  Not in Cross Save = its original membership type. Cross Save Primary = Any membership types it is overridding, and its original membership type Cross Save Overridden = Empty list
        bungie_global_display_name: The bungie global display name, if set.
        bungie_global_display_name_code: The bungie global display name code, if set.
        cross_save_override: If there is a cross save override in effect, this value will tell you the type that is overridding this one.
        display_name: Display Name the player has chosen for themselves. The display name is optional when the data type is used as input to a platform API.
        icon_path: URL the Icon if available.
        is_public: If True, this is a public user membership.
        membership_id: Membership ID as they user is known in the Accounts service
        membership_type: Type of the membership. Not necessarily the native type.
        supplemental_display_name: A platform specific additional display name - ex: psn Real Name, bnet Unique Name, etc.
    """

    applicable_membership_types: list[Union["BungieMembershipType", int]] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    bungie_global_display_name: str = custom_field()
    bungie_global_display_name_code: int = custom_field()
    cross_save_override: Union["BungieMembershipType", int] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    display_name: str = custom_field()
    icon_path: str = custom_field()
    is_public: bool = custom_field()
    membership_id: int = custom_field()
    membership_type: Union["BungieMembershipType", int] = custom_field(converter=enum_converter("BungieMembershipType"))
    supplemental_display_name: str = custom_field()


@custom_define()
class GeneralUser(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        about: _No description given by bungie._
        blizzard_display_name: _No description given by bungie._
        cached_bungie_global_display_name: _No description given by bungie._
        cached_bungie_global_display_name_code: _No description given by bungie._
        context: _No description given by bungie._
        display_name: _No description given by bungie._
        fb_display_name: _No description given by bungie._
        first_access: _No description given by bungie._
        is_deleted: _No description given by bungie._
        last_ban_report_id: _No description given by bungie._
        last_update: _No description given by bungie._
        legacy_portal_u_i_d: _No description given by bungie._
        locale: _No description given by bungie._
        locale_inherit_default: _No description given by bungie._
        membership_id: _No description given by bungie._
        normalized_name: _No description given by bungie._
        profile_ban_expire: _No description given by bungie._
        profile_picture: _No description given by bungie._
        profile_picture_path: _No description given by bungie._
        profile_picture_wide_path: _No description given by bungie._
        profile_theme: _No description given by bungie._
        profile_theme_name: _No description given by bungie._
        psn_display_name: _No description given by bungie._
        show_activity: _No description given by bungie._
        show_group_messaging: _No description given by bungie._
        stadia_display_name: _No description given by bungie._
        status_date: _No description given by bungie._
        status_text: _No description given by bungie._
        steam_display_name: _No description given by bungie._
        success_message_flags: _No description given by bungie._
        twitch_display_name: _No description given by bungie._
        unique_name: _No description given by bungie._
        user_title: _No description given by bungie._
        user_title_display: _No description given by bungie._
        xbox_display_name: _No description given by bungie._
    """

    about: str = custom_field()
    blizzard_display_name: str = custom_field()
    cached_bungie_global_display_name: str = custom_field()
    cached_bungie_global_display_name_code: int = custom_field()
    context: "UserToUserContext" = custom_field()
    display_name: str = custom_field()
    fb_display_name: str = custom_field()
    first_access: datetime = custom_field()
    is_deleted: bool = custom_field()
    last_ban_report_id: int = custom_field()
    last_update: datetime = custom_field()
    legacy_portal_u_i_d: int = custom_field()
    locale: str = custom_field()
    locale_inherit_default: bool = custom_field()
    membership_id: int = custom_field()
    normalized_name: str = custom_field()
    profile_ban_expire: datetime = custom_field()
    profile_picture: int = custom_field()
    profile_picture_path: str = custom_field()
    profile_picture_wide_path: str = custom_field()
    profile_theme: int = custom_field()
    profile_theme_name: str = custom_field()
    psn_display_name: str = custom_field()
    show_activity: bool = custom_field()
    show_group_messaging: bool = custom_field()
    stadia_display_name: str = custom_field()
    status_date: datetime = custom_field()
    status_text: str = custom_field()
    steam_display_name: str = custom_field()
    success_message_flags: int = custom_field()
    twitch_display_name: str = custom_field()
    unique_name: str = custom_field()
    user_title: int = custom_field()
    user_title_display: str = custom_field()
    xbox_display_name: str = custom_field()


@custom_define()
class UserToUserContext(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        global_ignore_end_date: _No description given by bungie._
        ignore_status: _No description given by bungie._
        is_following: _No description given by bungie._
    """

    global_ignore_end_date: datetime = custom_field()
    ignore_status: "IgnoreResponse" = custom_field()
    is_following: bool = custom_field()


@custom_define()
class UserMembershipData(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        bungie_net_user: _No description given by bungie._
        destiny_memberships: this allows you to see destiny memberships that are visible and linked to this account (regardless of whether or not they have characters on the world server)
        primary_membership_id: If this property is populated, it will have the membership ID of the account considered to be "primary" in this user's cross save relationship.  If null, this user has no cross save relationship, nor primary account.
    """

    bungie_net_user: "GeneralUser" = custom_field()
    destiny_memberships: list["GroupUserInfoCard"] = custom_field(metadata={"type": """list[GroupUserInfoCard]"""})
    primary_membership_id: int = custom_field()


@custom_define()
class HardLinkedUserMembership(BaseModel, DestinyUserMixin):
    """
    _No description given by bungie._

    None
    Attributes:
        cross_save_overridden_membership_id: _No description given by bungie._
        cross_save_overridden_type: _No description given by bungie._
        membership_id: _No description given by bungie._
        membership_type: _No description given by bungie._
    """

    cross_save_overridden_membership_id: int = custom_field()
    cross_save_overridden_type: Union["BungieMembershipType", int] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    membership_id: int = custom_field()
    membership_type: Union["BungieMembershipType", int] = custom_field(converter=enum_converter("BungieMembershipType"))


@custom_define()
class UserSearchResponse(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        has_more: _No description given by bungie._
        page: _No description given by bungie._
        search_results: _No description given by bungie._
    """

    has_more: bool = custom_field()
    page: int = custom_field()
    search_results: list["UserSearchResponseDetail"] = custom_field(
        metadata={"type": """list[UserSearchResponseDetail]"""}
    )


@custom_define()
class UserSearchResponseDetail(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        bungie_global_display_name: _No description given by bungie._
        bungie_global_display_name_code: _No description given by bungie._
        bungie_net_membership_id: _No description given by bungie._
        destiny_memberships: _No description given by bungie._
    """

    bungie_global_display_name: str = custom_field()
    bungie_global_display_name_code: int = custom_field()
    bungie_net_membership_id: int = custom_field()
    destiny_memberships: list["UserInfoCard"] = custom_field(metadata={"type": """list[UserInfoCard]"""})


@custom_define()
class UserSearchPrefixRequest(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        display_name_prefix: _No description given by bungie._
    """

    display_name_prefix: str = custom_field()


@custom_define()
class ExactSearchRequest(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        display_name: _No description given by bungie._
        display_name_code: _No description given by bungie._
    """

    display_name: str = custom_field()
    display_name_code: int = custom_field()


@custom_define()
class EmailSettings(BaseModel):
    """
    The set of all email subscription/opt-in settings and definitions.

    None
    Attributes:
        opt_in_definitions: Keyed by the name identifier of the opt-in definition.
        subscription_definitions: Keyed by the name identifier of the Subscription definition.
        views: Keyed by the name identifier of the View definition.
    """

    opt_in_definitions: dict[str, "EmailOptInDefinition"] = custom_field(
        metadata={"type": """dict[str, EmailOptInDefinition]"""}
    )
    subscription_definitions: dict[str, "EmailSubscriptionDefinition"] = custom_field(
        metadata={"type": """dict[str, EmailSubscriptionDefinition]"""}
    )
    views: dict[str, "EmailViewDefinition"] = custom_field(metadata={"type": """dict[str, EmailViewDefinition]"""})


@custom_define()
class EmailOptInDefinition(BaseModel):
    """
    Defines a single opt-in category: a wide-scoped permission to send emails for the subject related to the opt-in.

    None
    Attributes:
        dependent_subscriptions: Information about the dependent subscriptions for this opt-in.
        name: The unique identifier for this opt-in category.
        set_by_default: If true, this opt-in setting should be set by default in situations where accounts are created without explicit choices about what they're opting into.
        value: The flag value for this opt-in category. For historical reasons, this is defined as a flags enum.
    """

    dependent_subscriptions: list["EmailSubscriptionDefinition"] = custom_field(
        metadata={"type": """list[EmailSubscriptionDefinition]"""}
    )
    name: str = custom_field()
    set_by_default: bool = custom_field()
    value: Union["OptInFlags", int] = custom_field(converter=enum_converter("OptInFlags"))


class OptInFlags(BaseFlagEnum):
    """
    _No description given by bungie._
    """

    NONE = 0
    """_No description given by bungie._ """
    NEWSLETTER = 1
    """_No description given by bungie._ """
    SYSTEM = 2
    """_No description given by bungie._ """
    MARKETING = 4
    """_No description given by bungie._ """
    USER_RESEARCH = 8
    """_No description given by bungie._ """
    CUSTOMER_SERVICE = 16
    """_No description given by bungie._ """
    SOCIAL = 32
    """_No description given by bungie._ """
    PLAY_TESTS = 64
    """_No description given by bungie._ """
    PLAY_TESTS_LOCAL = 128
    """_No description given by bungie._ """
    CAREERS = 256
    """_No description given by bungie._ """


@custom_define()
class EmailSubscriptionDefinition(BaseModel):
    """
    Defines a single subscription: permission to send emails for a specific, focused subject (generally timeboxed, such as for a specific release of a product or feature).

    None
    Attributes:
        localization: A dictionary of localized text for the EMail Opt-in setting, keyed by the locale.
        name: The unique identifier for this subscription.
        value: The bitflag value for this subscription. Should be a unique power of two value.
    """

    localization: dict[str, "EMailSettingSubscriptionLocalization"] = custom_field(
        metadata={"type": """dict[str, EMailSettingSubscriptionLocalization]"""}
    )
    name: str = custom_field()
    value: int = custom_field()


@custom_define()
class EMailSettingLocalization(BaseModel):
    """
    Localized text relevant to a given EMail setting in a given localization.

    None
    Attributes:
        description: _No description given by bungie._
        title: _No description given by bungie._
    """

    description: str = custom_field()
    title: str = custom_field()


@custom_define()
class EMailSettingSubscriptionLocalization(BaseModel):
    """
    Localized text relevant to a given EMail setting in a given localization. Extra settings specifically for subscriptions.

    None
    Attributes:
        description: _No description given by bungie._
        known_user_action_text: _No description given by bungie._
        registered_user_description: _No description given by bungie._
        title: _No description given by bungie._
        unknown_user_action_text: _No description given by bungie._
        unknown_user_description: _No description given by bungie._
        unregistered_user_description: _No description given by bungie._
    """

    description: str = custom_field()
    known_user_action_text: str = custom_field()
    registered_user_description: str = custom_field()
    title: str = custom_field()
    unknown_user_action_text: str = custom_field()
    unknown_user_description: str = custom_field()
    unregistered_user_description: str = custom_field()


@custom_define()
class EmailViewDefinition(BaseModel):
    """
    Represents a data-driven view for Email settings. Web/Mobile UI can use this data to show new EMail settings consistently without further manual work.

    None
    Attributes:
        name: The identifier for this view.
        view_settings: The ordered list of settings to show in this view.
    """

    name: str = custom_field()
    view_settings: list["EmailViewDefinitionSetting"] = custom_field(
        metadata={"type": """list[EmailViewDefinitionSetting]"""}
    )


@custom_define()
class EmailViewDefinitionSetting(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        localization: A dictionary of localized text for the EMail setting, keyed by the locale.
        name: The identifier for this UI Setting, which can be used to relate it to custom strings or other data as desired.
        opt_in_aggregate_value: The OptInFlags value to set or clear if this setting is set or cleared in the UI. It is the aggregate of all underlying opt-in flags related to this setting.
        set_by_default: If true, this setting should be set by default if the user hasn't chosen whether it's set or cleared yet.
        subscriptions: The subscriptions to show as children of this setting, if any.
    """

    localization: dict[str, "EMailSettingLocalization"] = custom_field(
        metadata={"type": """dict[str, EMailSettingLocalization]"""}
    )
    name: str = custom_field()
    opt_in_aggregate_value: Union["OptInFlags", int] = custom_field(converter=enum_converter("OptInFlags"))
    set_by_default: bool = custom_field()
    subscriptions: list["EmailSubscriptionDefinition"] = custom_field(
        metadata={"type": """list[EmailSubscriptionDefinition]"""}
    )



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import TYPE_CHECKING, Optional, Union

from bungio.models.base import BaseModel, ManifestModel, custom_define, custom_field
from bungio.utils import enum_converter

if TYPE_CHECKING:
    from bungio.models import (
        DestinyCollectibleDefinition,
        DestinyDisplayPropertiesDefinition,
        DestinyInventoryItemDefinition,
        DestinyMetricDefinition,
        DestinyObjectiveDefinition,
        DestinyPresentationDisplayStyle,
        DestinyPresentationNodeType,
        DestinyPresentationScreenStyle,
        DestinyRecordDefinition,
        DestinyScope,
    )


@custom_define()
class DestinyPresentationNodeBaseDefinition(BaseModel):
    """
    This is the base class for all presentation system children. Presentation Nodes, Records, Collectibles, and Metrics.

    None
    Attributes:
        hash: The unique identifier for this entity. Guaranteed to be unique for the type of entity, but not globally. When entities refer to each other in Destiny content, it is this hash that they are referring to.
        index: The index of the entity as it was found in the investment tables.
        parent_node_hashes: A quick reference to presentation nodes that have this node as a child. Presentation nodes can be parented under multiple parents.
        presentation_node_type: _No description given by bungie._
        redacted: If this is true, then there is an entity with this identifier/type combination, but BNet is not yet allowed to show it. Sorry!
        trait_hashes: _No description given by bungie._
        trait_ids: _No description given by bungie._
    """

    hash: int = custom_field()
    index: int = custom_field()
    parent_node_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    presentation_node_type: Union["DestinyPresentationNodeType", int] = custom_field(
        converter=enum_converter("DestinyPresentationNodeType")
    )
    redacted: bool = custom_field()
    trait_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    trait_ids: list[str] = custom_field(metadata={"type": """list[str]"""})


@custom_define()
class DestinyScoredPresentationNodeBaseDefinition(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        hash: The unique identifier for this entity. Guaranteed to be unique for the type of entity, but not globally. When entities refer to each other in Destiny content, it is this hash that they are referring to.
        index: The index of the entity as it was found in the investment tables.
        max_category_record_score: _No description given by bungie._
        parent_node_hashes: A quick reference to presentation nodes that have this node as a child. Presentation nodes can be parented under multiple parents.
        presentation_node_type: _No description given by bungie._
        redacted: If this is true, then there is an entity with this identifier/type combination, but BNet is not yet allowed to show it. Sorry!
        trait_hashes: _No description given by bungie._
        trait_ids: _No description given by bungie._
    """

    hash: int = custom_field()
    index: int = custom_field()
    max_category_record_score: int = custom_field()
    parent_node_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    presentation_node_type: Union["DestinyPresentationNodeType", int] = custom_field(
        converter=enum_converter("DestinyPresentationNodeType")
    )
    redacted: bool = custom_field()
    trait_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    trait_ids: list[str] = custom_field(metadata={"type": """list[str]"""})


@custom_define()
class DestinyPresentationNodeDefinition(ManifestModel):
    """
    A PresentationNode is an entity that represents a logical grouping of other entities visually/organizationally. For now, Presentation Nodes may contain the following... but it may be used for more in the future: - Collectibles - Records (Or, as the public will call them, "Triumphs." Don't ask me why we're overloading the term "Triumph", it still hurts me to think about it) - Metrics (aka Stat Trackers) - Other Presentation Nodes, allowing a tree of Presentation Nodes to be created Part of me wants to break these into conceptual definitions per entity being collected, but the possibility of these different types being mixed in the same UI and the possibility that it could actually be more useful to return the "bare metal" presentation node concept has resulted in me deciding against that for the time being. We'll see if I come to regret this as well.

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        children: The child entities contained by this presentation node.
        completion_record_hash: If this presentation node has an associated "Record" that you can accomplish for completing its children, this is the identifier of that Record.
        disable_child_subscreen_navigation: If this presentation node has children, but the game doesn't let you inspect the details of those children, that is indicated here.
        display_properties: _No description given by bungie._
        display_style: A hint for how to display this presentation node when it's shown in a list.
        hash: The unique identifier for this entity. Guaranteed to be unique for the type of entity, but not globally. When entities refer to each other in Destiny content, it is this hash that they are referring to.
        index: The index of the entity as it was found in the investment tables.
        max_category_record_score: _No description given by bungie._
        node_type: _No description given by bungie._
        objective_hash: If this presentation node shows a related objective (for instance, if it tracks the progress of its children), the objective being tracked is indicated here.
        original_icon: The original icon for this presentation node, before we futzed with it.
        parent_node_hashes: A quick reference to presentation nodes that have this node as a child. Presentation nodes can be parented under multiple parents.
        presentation_node_type: _No description given by bungie._
        redacted: If this is true, then there is an entity with this identifier/type combination, but BNet is not yet allowed to show it. Sorry!
        requirements: The requirements for being able to interact with this presentation node and its children.
        root_view_icon: Some presentation nodes are meant to be explicitly shown on the "root" or "entry" screens for the feature to which they are related. You should use this icon when showing them on such a view, if you have a similar "entry point" view in your UI. If you don't have a UI, then I guess it doesn't matter either way does it?
        scope: Indicates whether this presentation node's state is determined on a per-character or on an account-wide basis.
        screen_style: A hint for how to display this presentation node when it's shown in its own detail screen.
        trait_hashes: _No description given by bungie._
        trait_ids: _No description given by bungie._
        manifest_completion_record_hash: Manifest information for `completion_record_hash`
        manifest_objective_hash: Manifest information for `objective_hash`
    """

    children: "DestinyPresentationNodeChildrenBlock" = custom_field()
    completion_record_hash: int = custom_field()
    disable_child_subscreen_navigation: bool = custom_field()
    display_properties: "DestinyDisplayPropertiesDefinition" = custom_field()
    display_style: Union["DestinyPresentationDisplayStyle", int] = custom_field(
        converter=enum_converter("DestinyPresentationDisplayStyle")
    )
    hash: int = custom_field()
    index: int = custom_field()
    max_category_record_score: int = custom_field()
    node_type: Union["DestinyPresentationNodeType", int] = custom_field(
        converter=enum_converter("DestinyPresentationNodeType")
    )
    objective_hash: int = custom_field()
    original_icon: str = custom_field()
    parent_node_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    presentation_node_type: Union["DestinyPresentationNodeType", int] = custom_field(
        converter=enum_converter("DestinyPresentationNodeType")
    )
    redacted: bool = custom_field()
    requirements: "DestinyPresentationNodeRequirementsBlock" = custom_field()
    root_view_icon: str = custom_field()
    scope: Union["DestinyScope", int] = custom_field(converter=enum_converter("DestinyScope"))
    screen_style: Union["DestinyPresentationScreenStyle", int] = custom_field(
        converter=enum_converter("DestinyPresentationScreenStyle")
    )
    trait_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    trait_ids: list[str] = custom_field(metadata={"type": """list[str]"""})
    manifest_completion_record_hash: Optional["DestinyRecordDefinition"] = custom_field(default=None)
    manifest_objective_hash: Optional["DestinyObjectiveDefinition"] = custom_field(default=None)


@custom_define()
class DestinyPresentationNodeChildrenBlock(BaseModel):
    """
    As/if presentation nodes begin to host more entities as children, these lists will be added to. One list property exists per type of entity that can be treated as a child of this presentation node, and each holds the identifier of the entity and any associated information needed to display the UI for that entity (if anything)

    None
    Attributes:
        collectibles: _No description given by bungie._
        craftables: _No description given by bungie._
        metrics: _No description given by bungie._
        presentation_nodes: _No description given by bungie._
        records: _No description given by bungie._
    """

    collectibles: list["DestinyPresentationNodeCollectibleChildEntry"] = custom_field(
        metadata={"type": """list[DestinyPresentationNodeCollectibleChildEntry]"""}
    )
    craftables: list["DestinyPresentationNodeCraftableChildEntry"] = custom_field(
        metadata={"type": """list[DestinyPresentationNodeCraftableChildEntry]"""}
    )
    metrics: list["DestinyPresentationNodeMetricChildEntry"] = custom_field(
        metadata={"type": """list[DestinyPresentationNodeMetricChildEntry]"""}
    )
    presentation_nodes: list["DestinyPresentationNodeChildEntry"] = custom_field(
        metadata={"type": """list[DestinyPresentationNodeChildEntry]"""}
    )
    records: list["DestinyPresentationNodeRecordChildEntry"] = custom_field(
        metadata={"type": """list[DestinyPresentationNodeRecordChildEntry]"""}
    )


@custom_define()
class DestinyPresentationNodeChildEntryBase(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        node_display_priority: Use this value to sort the presentation node children in ascending order.
    """

    node_display_priority: int = custom_field()


@custom_define()
class DestinyPresentationNodeChildEntry(BaseModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        node_display_priority: Use this value to sort the presentation node children in ascending order.
        presentation_node_hash: _No description given by bungie._
        manifest_presentation_node_hash: Manifest information for `presentation_node_hash`
    """

    node_display_priority: int = custom_field()
    presentation_node_hash: int = custom_field()
    manifest_presentation_node_hash: Optional["DestinyPresentationNodeDefinition"] = custom_field(default=None)


@custom_define()
class DestinyPresentationNodeCollectibleChildEntry(BaseModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        collectible_hash: _No description given by bungie._
        node_display_priority: Use this value to sort the presentation node children in ascending order.
        manifest_collectible_hash: Manifest information for `collectible_hash`
    """

    collectible_hash: int = custom_field()
    node_display_priority: int = custom_field()
    manifest_collectible_hash: Optional["DestinyCollectibleDefinition"] = custom_field(default=None)


@custom_define()
class DestinyPresentationNodeRequirementsBlock(BaseModel):
    """
    Presentation nodes can be restricted by various requirements. This defines the rules of those requirements, and the message(s) to be shown if these requirements aren't met.

    None
    Attributes:
        entitlement_unavailable_message: If this node is not accessible due to Entitlements (for instance, you don't own the required game expansion), this is the message to show.
    """

    entitlement_unavailable_message: str = custom_field()


@custom_define()
class DestinyPresentationChildBlock(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        display_style: _No description given by bungie._
        parent_presentation_node_hashes: _No description given by bungie._
        presentation_node_type: _No description given by bungie._
    """

    display_style: Union["DestinyPresentationDisplayStyle", int] = custom_field(
        converter=enum_converter("DestinyPresentationDisplayStyle")
    )
    parent_presentation_node_hashes: list[int] = custom_field(metadata={"type": """list[int]"""})
    presentation_node_type: Union["DestinyPresentationNodeType", int] = custom_field(
        converter=enum_converter("DestinyPresentationNodeType")
    )


@custom_define()
class DestinyPresentationNodeRecordChildEntry(BaseModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        node_display_priority: Use this value to sort the presentation node children in ascending order.
        record_hash: _No description given by bungie._
        manifest_record_hash: Manifest information for `record_hash`
    """

    node_display_priority: int = custom_field()
    record_hash: int = custom_field()
    manifest_record_hash: Optional["DestinyRecordDefinition"] = custom_field(default=None)


@custom_define()
class DestinyPresentationNodeMetricChildEntry(BaseModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        metric_hash: _No description given by bungie._
        node_display_priority: Use this value to sort the presentation node children in ascending order.
        manifest_metric_hash: Manifest information for `metric_hash`
    """

    metric_hash: int = custom_field()
    node_display_priority: int = custom_field()
    manifest_metric_hash: Optional["DestinyMetricDefinition"] = custom_field(default=None)


@custom_define()
class DestinyPresentationNodeCraftableChildEntry(BaseModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        craftable_item_hash: _No description given by bungie._
        node_display_priority: Use this value to sort the presentation node children in ascending order.
        manifest_craftable_item_hash: Manifest information for `craftable_item_hash`
    """

    craftable_item_hash: int = custom_field()
    node_display_priority: int = custom_field()
    manifest_craftable_item_hash: Optional["DestinyInventoryItemDefinition"] = custom_field(default=None)



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

from bungio.models.base import BaseModel, custom_define, custom_field
from bungio.utils import enum_converter

if TYPE_CHECKING:
    from bungio.models import (
        DamageType,
        DestinyBreakerTypeDefinition,
        DestinyDamageTypeDefinition,
        DestinyEnergyType,
        DestinyEnergyTypeDefinition,
        DestinyInventoryBucketDefinition,
        DestinyInventoryItemDefinition,
        DestinyMetricDefinition,
        DestinyObjectiveProgress,
        DestinyPerkReference,
        DestinyProgression,
        DestinyStat,
        DestinyTalentGridDefinition,
        DestinyTalentNode,
        EquipFailureReason,
        ItemBindStatus,
        ItemLocation,
        ItemState,
        TransferStatuses,
    )


@custom_define()
class DestinyItemComponent(BaseModel):
    """
    The base item component, filled with properties that are generally useful to know in any item request or that don't feel worthwhile to put in their own component.

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        bind_status: If the item is bound to a location, it will be specified in this enum.
        bucket_hash: The hash identifier for the specific inventory bucket in which the item is located.
        expiration_date: If the item can expire, this is the date at which it will/did expire.
        is_wrapper: If this is true, the object is actually a "wrapper" of the object it's representing. This means that it's not the actual item itself, but rather an item that must be "opened" in game before you have and can use the item.  Wrappers are an evolution of "bundles", which give an easy way to let you preview the contents of what you purchased while still letting you get a refund before you "open" it.
        item_hash: The identifier for the item's definition, which is where most of the useful static information for the item can be found.
        item_instance_id: If the item is instanced, it will have an instance ID. Lack of an instance ID implies that the item has no distinct local qualities aside from stack size.
        item_value_visibility: If available, a list that describes which item values (rewards) should be shown (true) or hidden (false).
        location: An easy reference for where the item is located. Redundant if you got the item from an Inventory, but useful when making detail calls on specific items.
        lockable: If the item can be locked, this will indicate that state.
        metric_hash: The identifier for the currently-selected metric definition, to be displayed on the emblem nameplate.
        metric_objective: The objective progress for the currently-selected metric definition, to be displayed on the emblem nameplate.
        override_style_item_hash: If populated, this is the hash of the item whose icon (and other secondary styles, but *not* the human readable strings) should override whatever icons/styles are on the item being sold. If you don't do this, certain items whose styles are being overridden by socketed items - such as the "Recycle Shader" item - would show whatever their default icon/style is, and it wouldn't be pretty or look accurate.
        quantity: The quantity of the item in this stack. Note that Instanced items cannot stack. If an instanced item, this value will always be 1 (as the stack has exactly one item in it)
        state: A flags enumeration indicating the transient/custom states of the item that affect how it is rendered: whether it's tracked or locked for example, or whether it has a masterwork plug inserted.
        tooltip_notification_indexes: If this is populated, it is a list of indexes into DestinyInventoryItemDefinition.tooltipNotifications for any special tooltip messages that need to be shown for this item.
        transfer_status: If there is a known error state that would cause this item to not be transferable, this Flags enum will indicate all of those error states. Otherwise, it will be 0 (CanTransfer).
        version_number: The version of this item, used to index into the versions list in the item definition quality block.
        manifest_bucket_hash: Manifest information for `bucket_hash`
        manifest_item_hash: Manifest information for `item_hash`
        manifest_metric_hash: Manifest information for `metric_hash`
        manifest_override_style_item_hash: Manifest information for `override_style_item_hash`
    """

    bind_status: Union["ItemBindStatus", int] = custom_field(converter=enum_converter("ItemBindStatus"))
    bucket_hash: int = custom_field()
    expiration_date: datetime = custom_field()
    is_wrapper: bool = custom_field()
    item_hash: int = custom_field()
    item_instance_id: int = custom_field()
    item_value_visibility: list[bool] = custom_field(metadata={"type": """list[bool]"""})
    location: Union["ItemLocation", int] = custom_field(converter=enum_converter("ItemLocation"))
    lockable: bool = custom_field()
    metric_hash: int = custom_field()
    metric_objective: "DestinyObjectiveProgress" = custom_field()
    override_style_item_hash: int = custom_field()
    quantity: int = custom_field()
    state: Union["ItemState", int] = custom_field(converter=enum_converter("ItemState"))
    tooltip_notification_indexes: list[int] = custom_field(metadata={"type": """list[int]"""})
    transfer_status: Union["TransferStatuses", int] = custom_field(converter=enum_converter("TransferStatuses"))
    version_number: int = custom_field()
    manifest_bucket_hash: Optional["DestinyInventoryBucketDefinition"] = custom_field(default=None)
    manifest_item_hash: Optional["DestinyInventoryItemDefinition"] = custom_field(default=None)
    manifest_metric_hash: Optional["DestinyMetricDefinition"] = custom_field(default=None)
    manifest_override_style_item_hash: Optional["DestinyInventoryItemDefinition"] = custom_field(default=None)


@custom_define()
class DestinyItemPerksComponent(BaseModel):
    """
    Instanced items can have perks: benefits that the item bestows. These are related to DestinySandboxPerkDefinition, and sometimes - but not always - have human readable info. When they do, they are the icons and text that you see in an item's tooltip. Talent Grids, Sockets, and the item itself can apply Perks, which are then summarized here for your convenience.

    None
    Attributes:
        perks: The list of perks to display in an item tooltip - and whether or not they have been activated.
    """

    perks: list["DestinyPerkReference"] = custom_field(metadata={"type": """list[DestinyPerkReference]"""})


@custom_define()
class DestinyItemObjectivesComponent(BaseModel):
    """
    Items can have objectives and progression. When you request this block, you will obtain information about any Objectives and progression tied to this item.

    None
    Attributes:
        date_completed: If we have any information on when these objectives were completed, this will be the date of that completion. This won't be on many items, but could be interesting for some items that do store this information.
        flavor_objective: I may regret naming it this way - but this represents when an item has an objective that doesn't serve a beneficial purpose, but rather is used for "flavor" or additional information. For instance, when Emblems track specific stats, those stats are represented as Objectives on the item.
        objectives: If the item has a hard association with objectives, your progress on them will be defined here.  Objectives are our standard way to describe a series of tasks that have to be completed for a reward.
    """

    date_completed: datetime = custom_field()
    flavor_objective: "DestinyObjectiveProgress" = custom_field()
    objectives: list["DestinyObjectiveProgress"] = custom_field(metadata={"type": """list[DestinyObjectiveProgress]"""})


@custom_define()
class DestinyItemInstanceComponent(BaseModel):
    """
    If an item is "instanced", this will contain information about the item's instance that doesn't fit easily into other components. One might say this is the "essential" instance data for the item. Items are instanced if they require information or state that can vary. For instance, weapons are Instanced: they are given a unique identifier, uniquely generated stats, and can have their properties altered. Non-instanced items have none of these things: for instance, Glimmer has no unique properties aside from how much of it you own. You can tell from an item's definition whether it will be instanced or not by looking at the DestinyInventoryItemDefinition's definition.inventory.isInstanceItem property.

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        breaker_type: If populated, this item has a breaker type corresponding to the given value. See DestinyBreakerTypeDefinition for more details.
        breaker_type_hash: If populated, this is the hash identifier for the item's breaker type. See DestinyBreakerTypeDefinition for more details.
        can_equip: If this is an equippable item, you can check it here. There are permanent as well as transitory reasons why an item might not be able to be equipped: check cannotEquipReason for details.
        cannot_equip_reason: If you cannot equip the item, this is a flags enum that enumerates all of the reasons why you couldn't equip the item. You may need to refine your UI further by using unlockHashesRequiredToEquip and equipRequiredLevel.
        damage_type: If the item has a damage type, this is the item's current damage type.
        damage_type_hash: The current damage type's hash, so you can look up localized info and icons for it.
        energy: IF populated, this item supports Energy mechanics (i.e. Armor 2.0), and these are the current details of its energy type and available capacity to spend energy points.
        equip_required_level: If the item cannot be equipped until you reach a certain level, that level will be reflected here.
        is_equipped: Is the item currently equipped on the given character?
        item_level: The Item's "Level" has the most significant bearing on its stats, such as Light and Power.
        primary_stat: The item stat that we consider to be "primary" for the item. For instance, this would be "Attack" for Weapons or "Defense" for armor.
        quality: The "Quality" of the item has a lesser - but still impactful - bearing on stats like Light and Power.
        unlock_hashes_required_to_equip: Sometimes, there are limitations to equipping that are represented by character-level flags called "unlocks". This is a list of flags that they need in order to equip the item that the character has not met. Use these to look up the descriptions to show in your UI by looking up the relevant DestinyUnlockDefinitions for the hashes.
        manifest_breaker_type_hash: Manifest information for `breaker_type_hash`
        manifest_damage_type_hash: Manifest information for `damage_type_hash`
    """

    breaker_type: int = custom_field()
    breaker_type_hash: int = custom_field()
    can_equip: bool = custom_field()
    cannot_equip_reason: Union["EquipFailureReason", int] = custom_field(converter=enum_converter("EquipFailureReason"))
    damage_type: Union["DamageType", int] = custom_field(converter=enum_converter("DamageType"))
    damage_type_hash: int = custom_field()
    energy: "DestinyItemInstanceEnergy" = custom_field()
    equip_required_level: int = custom_field()
    is_equipped: bool = custom_field()
    item_level: int = custom_field()
    primary_stat: "DestinyStat" = custom_field()
    quality: int = custom_field()
    unlock_hashes_required_to_equip: list[int] = custom_field(metadata={"type": """list[int]"""})
    manifest_breaker_type_hash: Optional["DestinyBreakerTypeDefinition"] = custom_field(default=None)
    manifest_damage_type_hash: Optional["DestinyDamageTypeDefinition"] = custom_field(default=None)


@custom_define()
class DestinyItemInstanceEnergy(BaseModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        energy_capacity: The total capacity of Energy that the item currently has, regardless of if it is currently being used.
        energy_type: This is the enum version of the Energy Type value, for convenience.
        energy_type_hash: The type of energy for this item. Plugs that require Energy can only be inserted if they have the "Any" Energy Type or the matching energy type of this item. This is a reference to the DestinyEnergyTypeDefinition for the energy type, where you can find extended info about it.
        energy_unused: The amount of energy still available for inserting new plugs.
        energy_used: The amount of Energy currently in use by inserted plugs.
        manifest_energy_type_hash: Manifest information for `energy_type_hash`
    """

    energy_capacity: int = custom_field()
    energy_type: Union["DestinyEnergyType", int] = custom_field(converter=enum_converter("DestinyEnergyType"))
    energy_type_hash: int = custom_field()
    energy_unused: int = custom_field()
    energy_used: int = custom_field()
    manifest_energy_type_hash: Optional["DestinyEnergyTypeDefinition"] = custom_field(default=None)


@custom_define()
class DestinyItemRenderComponent(BaseModel):
    """
    Many items can be rendered in 3D. When you request this block, you will obtain the custom data needed to render this specific instance of the item.

    None
    Attributes:
        art_regions: A dictionary for rendering gear components, with: key = Art Arrangement Region Index value = The chosen Arrangement Index for the Region, based on the value of a stat on the item used for making the choice.
        use_custom_dyes: If you should use custom dyes on this item, it will be indicated here.
    """

    art_regions: dict[int, int] = custom_field(metadata={"type": """dict[int, int]"""})
    use_custom_dyes: bool = custom_field()


@custom_define()
class DestinyItemStatsComponent(BaseModel):
    """
    If you want the stats on an item's instanced data, get this component. These are stats like Attack, Defense etc... and *not* historical stats. Note that some stats have additional computation in-game at runtime - for instance, Magazine Size - and thus these stats might not be 100% accurate compared to what you see in-game for some stats. I know, it sucks. I hate it too.

    None
    Attributes:
        stats: If the item has stats that it provides (damage, defense, etc...), it will be given here.
    """

    stats: dict[int, "DestinyStat"] = custom_field(metadata={"type": """dict[int, DestinyStat]"""})


@custom_define()
class DestinyItemSocketsComponent(BaseModel):
    """
    Instanced items can have sockets, which are slots on the item where plugs can be inserted. Sockets are a bit complex: be sure to examine the documentation on the DestinyInventoryItemDefinition's "socket" block and elsewhere on these objects for more details.

    None
    Attributes:
        sockets: The list of all sockets on the item, and their status information.
    """

    sockets: list["DestinyItemSocketState"] = custom_field(metadata={"type": """list[DestinyItemSocketState]"""})


@custom_define()
class DestinyItemSocketState(BaseModel):
    """
    The status of a given item's socket. (which plug is inserted, if any: whether it is enabled, what "reusable" plugs can be inserted, etc...) If I had it to do over, this would probably have a DestinyItemPlug representing the inserted item instead of most of these properties. :shrug:

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        enable_fail_indexes: If a plug is inserted but not enabled, this will be populated with indexes into the plug item definition's plug.enabledRules property, so that you can show the reasons why it is not enabled.
        is_enabled: Even if a plug is inserted, it doesn't mean it's enabled. This flag indicates whether the plug is active and providing its benefits.
        is_visible: A plug may theoretically provide benefits but not be visible - for instance, some older items use a plug's damage type perk to modify their own damage type. These, though they are not visible, still affect the item. This field indicates that state. An invisible plug, while it provides benefits if it is Enabled, cannot be directly modified by the user.
        plug_hash: The currently active plug, if any. Note that, because all plugs are statically defined, its effect on stats and perks can be statically determined using the plug item's definition. The stats and perks can be taken at face value on the plug item as the stats and perks it will provide to the user/item.
        manifest_plug_hash: Manifest information for `plug_hash`
    """

    enable_fail_indexes: list[int] = custom_field(metadata={"type": """list[int]"""})
    is_enabled: bool = custom_field()
    is_visible: bool = custom_field()
    plug_hash: int = custom_field()
    manifest_plug_hash: Optional["DestinyInventoryItemDefinition"] = custom_field(default=None)


@custom_define()
class DestinyItemTalentGridComponent(BaseModel):
    """
    Well, we're here in Destiny 2, and Talent Grids are unfortunately still around. The good news is that they're pretty much only being used for certain base information on items and for Builds/Subclasses. The bad news is that they still suck. If you really want this information, grab this component. An important note is that talent grids are defined as such: A Grid has 1:M Nodes, which has 1:M Steps. Any given node can only have a single step active at one time, which represents the actual visual contents and effects of the Node (for instance, if you see a "Super Cool Bonus" node, the actual icon and text for the node is coming from the current Step of that node). Nodes can be grouped into exclusivity sets *and* as of D2, exclusivity groups (which are collections of exclusivity sets that affect each other). See DestinyTalentGridDefinition for more information. Brace yourself, the water's cold out there in the deep end.

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        grid_progression: If the item has a progression, it will be detailed here. A progression means that the item can gain experience. Thresholds of experience are what determines whether and when a talent node can be activated.
        is_grid_complete: Indicates whether the talent grid on this item is completed, and thus whether it should have a gold border around it. Only will be true if the item actually *has* a talent grid, and only then if it is completed (i.e. every exclusive set has an activated node, and every non-exclusive set node has been activated)
        nodes: Detailed information about the individual nodes in the talent grid. A node represents a single visual "pip" in the talent grid or Build detail view, though each node may have multiple "steps" which indicate the actual bonuses and visual representation of that node.
        talent_grid_hash: Most items don't have useful talent grids anymore, but Builds in particular still do. You can use this hash to lookup the DestinyTalentGridDefinition attached to this item, which will be crucial for understanding the node values on the item.
        manifest_talent_grid_hash: Manifest information for `talent_grid_hash`
    """

    grid_progression: "DestinyProgression" = custom_field()
    is_grid_complete: bool = custom_field()
    nodes: list["DestinyTalentNode"] = custom_field(metadata={"type": """list[DestinyTalentNode]"""})
    talent_grid_hash: int = custom_field()
    manifest_talent_grid_hash: Optional["DestinyTalentGridDefinition"] = custom_field(default=None)



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from datetime import datetime
from typing import TYPE_CHECKING, Union

from bungio.models.base import BaseEnum, BaseModel, custom_define, custom_field
from bungio.models.mixins import DestinyClanMixin, DestinyUserMixin
from bungio.utils import enum_converter

if TYPE_CHECKING:
    from bungio.models import BungieMembershipType, UserInfoCard


class FireteamDateRange(BaseEnum):
    """
    _No description given by bungie._
    """

    ALL = 0
    """_No description given by bungie._ """
    NOW = 1
    """_No description given by bungie._ """
    TWENTY_FOUR_HOURS = 2
    """_No description given by bungie._ """
    FORTY_EIGHT_HOURS = 3
    """_No description given by bungie._ """
    THIS_WEEK = 4
    """_No description given by bungie._ """


class FireteamPlatform(BaseEnum):
    """
    _No description given by bungie._
    """

    ANY = 0
    """_No description given by bungie._ """
    PLAYSTATION4 = 1
    """_No description given by bungie._ """
    XBOX_ONE = 2
    """_No description given by bungie._ """
    BLIZZARD = 3
    """_No description given by bungie._ """
    STEAM = 4
    """_No description given by bungie._ """
    STADIA = 5
    """_No description given by bungie._ """


class FireteamPublicSearchOption(BaseEnum):
    """
    _No description given by bungie._
    """

    PUBLIC_AND_PRIVATE = 0
    """_No description given by bungie._ """
    PUBLIC_ONLY = 1
    """_No description given by bungie._ """
    PRIVATE_ONLY = 2
    """_No description given by bungie._ """


class FireteamSlotSearch(BaseEnum):
    """
    _No description given by bungie._
    """

    NO_SLOT_RESTRICTION = 0
    """_No description given by bungie._ """
    HAS_OPEN_PLAYER_SLOTS = 1
    """_No description given by bungie._ """
    HAS_OPEN_PLAYER_OR_ALT_SLOTS = 2
    """_No description given by bungie._ """


@custom_define()
class FireteamSummary(BaseModel, DestinyClanMixin):
    """
    _No description given by bungie._

    None
    Attributes:
        activity_type: _No description given by bungie._
        alternate_slot_count: _No description given by bungie._
        available_alternate_slot_count: _No description given by bungie._
        available_player_slot_count: _No description given by bungie._
        date_created: _No description given by bungie._
        date_modified: _No description given by bungie._
        date_player_modified: _No description given by bungie._
        fireteam_id: _No description given by bungie._
        group_id: _No description given by bungie._
        is_immediate: _No description given by bungie._
        is_public: _No description given by bungie._
        is_valid: _No description given by bungie._
        locale: _No description given by bungie._
        owner_membership_id: _No description given by bungie._
        platform: _No description given by bungie._
        player_slot_count: _No description given by bungie._
        scheduled_time: _No description given by bungie._
        title: _No description given by bungie._
        title_before_moderation: _No description given by bungie._
    """

    activity_type: int = custom_field()
    alternate_slot_count: int = custom_field()
    available_alternate_slot_count: int = custom_field()
    available_player_slot_count: int = custom_field()
    date_created: datetime = custom_field()
    date_modified: datetime = custom_field()
    date_player_modified: datetime = custom_field()
    fireteam_id: int = custom_field()
    group_id: int = custom_field()
    is_immediate: bool = custom_field()
    is_public: bool = custom_field()
    is_valid: bool = custom_field()
    locale: str = custom_field()
    owner_membership_id: int = custom_field()
    platform: Union["FireteamPlatform", int] = custom_field(converter=enum_converter("FireteamPlatform"))
    player_slot_count: int = custom_field()
    scheduled_time: datetime = custom_field()
    title: str = custom_field()
    title_before_moderation: str = custom_field()


@custom_define()
class FireteamResponse(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        alternates: _No description given by bungie._
        members: _No description given by bungie._
        summary: _No description given by bungie._
    """

    alternates: list["FireteamMember"] = custom_field(metadata={"type": """list[FireteamMember]"""})
    members: list["FireteamMember"] = custom_field(metadata={"type": """list[FireteamMember]"""})
    summary: "FireteamSummary" = custom_field()


@custom_define()
class FireteamMember(BaseModel):
    """
    _No description given by bungie._

    None
    Attributes:
        bungie_net_user_info: _No description given by bungie._
        character_id: _No description given by bungie._
        date_joined: _No description given by bungie._
        destiny_user_info: _No description given by bungie._
        has_microphone: _No description given by bungie._
        last_platform_invite_attempt_date: _No description given by bungie._
        last_platform_invite_attempt_result: _No description given by bungie._
    """

    bungie_net_user_info: "UserInfoCard" = custom_field()
    character_id: int = custom_field()
    date_joined: datetime = custom_field()
    destiny_user_info: "FireteamUserInfoCard" = custom_field()
    has_microphone: bool = custom_field()
    last_platform_invite_attempt_date: datetime = custom_field()
    last_platform_invite_attempt_result: Union["FireteamPlatformInviteResult", int] = custom_field(
        converter=enum_converter("FireteamPlatformInviteResult")
    )


@custom_define()
class FireteamUserInfoCard(BaseModel, DestinyUserMixin):
    """
    _No description given by bungie._

    None
    Attributes:
        applicable_membership_types: The list of Membership Types indicating the platforms on which this Membership can be used.  Not in Cross Save = its original membership type. Cross Save Primary = Any membership types it is overridding, and its original membership type Cross Save Overridden = Empty list
        bungie_global_display_name: The bungie global display name, if set.
        bungie_global_display_name_code: The bungie global display name code, if set.
        cross_save_override: If there is a cross save override in effect, this value will tell you the type that is overridding this one.
        display_name: Display Name the player has chosen for themselves. The display name is optional when the data type is used as input to a platform API.
        fireteam_display_name: _No description given by bungie._
        fireteam_membership_type: _No description given by bungie._
        icon_path: URL the Icon if available.
        is_public: If True, this is a public user membership.
        membership_id: Membership ID as they user is known in the Accounts service
        membership_type: Type of the membership. Not necessarily the native type.
        supplemental_display_name: A platform specific additional display name - ex: psn Real Name, bnet Unique Name, etc.
    """

    applicable_membership_types: list[Union["BungieMembershipType", int]] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    bungie_global_display_name: str = custom_field()
    bungie_global_display_name_code: int = custom_field()
    cross_save_override: Union["BungieMembershipType", int] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    display_name: str = custom_field()
    fireteam_display_name: str = custom_field()
    fireteam_membership_type: Union["BungieMembershipType", int] = custom_field(
        converter=enum_converter("BungieMembershipType")
    )
    icon_path: str = custom_field()
    is_public: bool = custom_field()
    membership_id: int = custom_field()
    membership_type: Union["BungieMembershipType", int] = custom_field(converter=enum_converter("BungieMembershipType"))
    supplemental_display_name: str = custom_field()


class FireteamPlatformInviteResult(BaseEnum):
    """
    _No description given by bungie._
    """

    NONE = 0
    """_No description given by bungie._ """
    SUCCESS = 1
    """_No description given by bungie._ """
    ALREADY_IN_FIRETEAM = 2
    """_No description given by bungie._ """
    THROTTLED = 3
    """_No description given by bungie._ """
    SERVICE_ERROR = 4
    """_No description given by bungie._ """



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.


from bungio.models.base import ManifestModel, custom_define, custom_field


@custom_define()
class DestinyPowerCapDefinition(ManifestModel):
    """
    Defines a 'power cap' (limit) for gear items, based on the rarity tier and season of release.

    None
    Attributes:
        hash: The unique identifier for this entity. Guaranteed to be unique for the type of entity, but not globally. When entities refer to each other in Destiny content, it is this hash that they are referring to.
        index: The index of the entity as it was found in the investment tables.
        power_cap: The raw value for a power cap.
        redacted: If this is true, then there is an entity with this identifier/type combination, but BNet is not yet allowed to show it. Sorry!
    """

    hash: int = custom_field()
    index: int = custom_field()
    power_cap: int = custom_field()
    redacted: bool = custom_field()



import datetime
import importlib
from typing import Callable, Type

from bungio.models.base import MISSING, BaseEnum, UnknownEnumValue

__all__ = ("get_now_with_tz", "enum_converter")


def get_now_with_tz() -> datetime.datetime:
    """
    Returns the timezone aware current datetime

    Returns:
        Timezone aware datetime
    """

    return datetime.datetime.now(tz=datetime.timezone.utc)


def enum_converter(enum_name: str) -> Callable:
    """
    Return a callable that can be used in attr converters to convert values to enums

    Args:
        enum_name: The name of the enum the value has

    Returns:
        A callable converter
    """

    def converter(
        value: int | str | list[int | str],
    ) -> BaseEnum | UnknownEnumValue | list[BaseEnum | UnknownEnumValue]:
        if isinstance(value, list):
            return [converter(v) for v in value]

        elif isinstance(value, BaseEnum | UnknownEnumValue | type(MISSING)):
            return value

        imp = importlib.import_module("bungio.models")
        enum_class = getattr(imp, enum_name)
        return enum_class(value)

    return converter


# all this does is allow objs to be asynchronously iterable, which does "fix" this bug https://bugs.python.org/issue33346
class AllowAsyncIteration:
    def __init__(self, obj):
        self._it = iter(obj)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            value = next(self._it)
        except StopIteration:
            raise StopAsyncIteration
        return value



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import TYPE_CHECKING

from bungio.models.base import BaseModel, custom_define, custom_field

if TYPE_CHECKING:
    from bungio.models import DestinyItemComponent


@custom_define()
class DestinyInventoryComponent(BaseModel):
    """
    A list of minimal information for items in an inventory: be it a character's inventory, or a Profile's inventory. (Note that the Vault is a collection of inventory buckets in the Profile's inventory) Inventory Items returned here are in a flat list, but importantly they have a bucketHash property that indicates the specific inventory bucket that is holding them. These buckets constitute things like the separate sections of the Vault, the user's inventory slots, etc. See DestinyInventoryBucketDefinition for more info.

    None
    Attributes:
        items: The items in this inventory. If you care to bucket them, use the item's bucketHash property to group them.
    """

    items: list["DestinyItemComponent"] = custom_field(metadata={"type": """list[DestinyItemComponent]"""})



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import Optional, Union

from bungio.models import (
    BungieMembershipType,
    ClanBanner,
    EntityActionResult,
    GetGroupsForMemberResponse,
    GroupApplicationListRequest,
    GroupApplicationRequest,
    GroupApplicationResponse,
    GroupBanRequest,
    GroupDateRange,
    GroupEditAction,
    GroupMemberLeaveResult,
    GroupMembershipSearchResponse,
    GroupNameSearchRequest,
    GroupOptionalConversation,
    GroupOptionalConversationAddRequest,
    GroupOptionalConversationEditRequest,
    GroupOptionsEditAction,
    GroupPotentialMembershipSearchResponse,
    GroupPotentialMemberStatus,
    GroupQuery,
    GroupResponse,
    GroupSearchResponse,
    GroupsForMemberFilter,
    GroupTheme,
    GroupType,
    GroupV2Card,
    RuntimeGroupMemberType,
    SearchResultOfGroupBan,
    SearchResultOfGroupMember,
    SearchResultOfGroupMemberApplication,
)
from bungio.models.auth import AuthData
from bungio.models.base import ClientMixin, custom_define
from bungio.utils import AllowAsyncIteration


@custom_define()
class GroupV2RouteInterface(ClientMixin):
    async def get_available_avatars(self, auth: Optional[AuthData] = None) -> dict[int, str]:
        """
        Returns a list of all available group avatars for the signed-in user.

        Args:
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_available_avatars(auth=auth)
        return {key: value async for key, value in AllowAsyncIteration(response["Response"].items())}

    async def get_available_themes(self, auth: Optional[AuthData] = None) -> list[GroupTheme]:
        """
        Returns a list of all available group themes.

        Args:
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_available_themes(auth=auth)
        return [
            await GroupTheme.from_dict(data=value, client=self._client, auth=auth) for value in response["Response"]
        ]

    async def get_user_clan_invite_setting(self, m_type: Union[BungieMembershipType, int], auth: AuthData) -> bool:
        """
        Gets the state of the user's clan invite preferences for a particular membership type - true if they wish to be invited to clans, false otherwise.

        Warning: Requires Authentication.
            Required oauth2 scopes: ReadUserData

        Args:
            m_type: The Destiny membership type of the account we wish to access settings.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_user_clan_invite_setting(
            m_type=getattr(m_type, "value", m_type), auth=auth
        )
        return response["Response"]

    async def get_recommended_groups(
        self, create_date_range: Union[GroupDateRange, int], group_type: Union[GroupType, int], auth: AuthData
    ) -> list[GroupV2Card]:
        """
        Gets groups recommended for you based on the groups to whom those you follow belong.

        Warning: Requires Authentication.
            Required oauth2 scopes: ReadGroups

        Args:
            create_date_range: Requested range in which to pull recommended groups
            group_type: Type of groups requested
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_recommended_groups(
            create_date_range=getattr(create_date_range, "value", create_date_range),
            group_type=getattr(group_type, "value", group_type),
            auth=auth,
        )
        return [
            await GroupV2Card.from_dict(
                data=value, client=self._client, create_date_range=create_date_range, group_type=group_type, auth=auth
            )
            for value in response["Response"]
        ]

    async def group_search(self, data: GroupQuery, auth: Optional[AuthData] = None) -> GroupSearchResponse:
        """
        Search for Groups.

        Args:
            data: The required data for this request.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.group_search(auth=auth, **data.to_dict())
        return await GroupSearchResponse.from_dict(data=response, client=self._client, auth=auth)

    async def get_group(self, group_id: int, auth: Optional[AuthData] = None) -> GroupResponse:
        """
        Get information about a specific group of the given ID.

        Args:
            group_id: Requested group's id.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_group(group_id=group_id, auth=auth)
        return await GroupResponse.from_dict(data=response, client=self._client, group_id=group_id, auth=auth)

    async def get_group_by_name(
        self, group_name: str, group_type: Union[GroupType, int], auth: Optional[AuthData] = None
    ) -> GroupResponse:
        """
        Get information about a specific group with the given name and type.

        Args:
            group_name: Exact name of the group to find.
            group_type: Type of group to find.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_group_by_name(
            group_name=group_name, group_type=getattr(group_type, "value", group_type), auth=auth
        )
        return await GroupResponse.from_dict(
            data=response, client=self._client, group_name=group_name, group_type=group_type, auth=auth
        )

    async def get_group_by_name_v2(
        self, data: GroupNameSearchRequest, auth: Optional[AuthData] = None
    ) -> GroupResponse:
        """
        Get information about a specific group with the given name and type. The POST version.

        Args:
            data: The required data for this request.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_group_by_name_v2(auth=auth, **data.to_dict())
        return await GroupResponse.from_dict(data=response, client=self._client, auth=auth)

    async def get_group_optional_conversations(
        self, group_id: int, auth: Optional[AuthData] = None
    ) -> list[GroupOptionalConversation]:
        """
        Gets a list of available optional conversation channels and their settings.

        Args:
            group_id: Requested group's id.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_group_optional_conversations(group_id=group_id, auth=auth)
        return [
            await GroupOptionalConversation.from_dict(data=value, client=self._client, group_id=group_id, auth=auth)
            for value in response["Response"]
        ]

    async def edit_group(self, data: GroupEditAction, group_id: int, auth: AuthData) -> int:
        """
        Edit an existing group. You must have suitable permissions in the group to perform this operation. This latest revision will only edit the fields you pass in - pass null for properties you want to leave unaltered.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: Group ID of the group to edit.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.edit_group(group_id=group_id, auth=auth, **data.to_dict())
        return response["Response"]

    async def edit_clan_banner(self, data: ClanBanner, group_id: int, auth: AuthData) -> int:
        """
        Edit an existing group's clan banner. You must have suitable permissions in the group to perform this operation. All fields are required.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: Group ID of the group to edit.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.edit_clan_banner(group_id=group_id, auth=auth, **data.to_dict())
        return response["Response"]

    async def edit_founder_options(self, data: GroupOptionsEditAction, group_id: int, auth: AuthData) -> int:
        """
        Edit group options only available to a founder. You must have suitable permissions in the group to perform this operation.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: Group ID of the group to edit.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.edit_founder_options(group_id=group_id, auth=auth, **data.to_dict())
        return response["Response"]

    async def add_optional_conversation(
        self, data: GroupOptionalConversationAddRequest, group_id: int, auth: AuthData
    ) -> int:
        """
        Add a new optional conversation/chat channel. Requires admin permissions to the group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: Group ID of the group to edit.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.add_optional_conversation(group_id=group_id, auth=auth, **data.to_dict())
        return response["Response"]

    async def edit_optional_conversation(
        self, data: GroupOptionalConversationEditRequest, conversation_id: int, group_id: int, auth: AuthData
    ) -> int:
        """
        Edit the settings of an optional conversation/chat channel. Requires admin permissions to the group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            conversation_id: Conversation Id of the channel being edited.
            group_id: Group ID of the group to edit.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.edit_optional_conversation(
            conversation_id=conversation_id, group_id=group_id, auth=auth, **data.to_dict()
        )
        return response["Response"]

    async def get_members_of_group(
        self,
        currentpage: int,
        group_id: int,
        member_type: Optional[Union[RuntimeGroupMemberType, int]] = None,
        name_search: Optional[str] = None,
        auth: Optional[AuthData] = None,
    ) -> SearchResultOfGroupMember:
        """
        Get the list of members in a given group.

        Args:
            currentpage: Page number (starting with 1). Each page has a fixed size of 50 items per page.
            group_id: The ID of the group.
            member_type: Filter out other member types. Use None for all members.
            name_search: The name fragment upon which a search should be executed for members with matching display or unique names.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_members_of_group(
            currentpage=currentpage,
            group_id=group_id,
            member_type=getattr(member_type, "value", member_type) if member_type else None,
            name_search=name_search if name_search else None,
            auth=auth,
        )
        return await SearchResultOfGroupMember.from_dict(
            data=response,
            client=self._client,
            currentpage=currentpage,
            group_id=group_id,
            member_type=member_type,
            name_search=name_search,
            auth=auth,
        )

    async def get_admins_and_founder_of_group(
        self, currentpage: int, group_id: int, auth: Optional[AuthData] = None
    ) -> SearchResultOfGroupMember:
        """
        Get the list of members in a given group who are of admin level or higher.

        Args:
            currentpage: Page number (starting with 1). Each page has a fixed size of 50 items per page.
            group_id: The ID of the group.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_admins_and_founder_of_group(
            currentpage=currentpage, group_id=group_id, auth=auth
        )
        return await SearchResultOfGroupMember.from_dict(
            data=response, client=self._client, currentpage=currentpage, group_id=group_id, auth=auth
        )

    async def edit_group_membership(
        self,
        group_id: int,
        membership_id: int,
        membership_type: Union[BungieMembershipType, int],
        member_type: Union[RuntimeGroupMemberType, int],
        auth: AuthData,
    ) -> int:
        """
        Edit the membership type of a given member. You must have suitable permissions in the group to perform this operation.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            group_id: ID of the group to which the member belongs.
            membership_id: Membership ID to modify.
            membership_type: Membership type of the provide membership ID.
            member_type: New membertype for the specified member.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.edit_group_membership(
            group_id=group_id,
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            member_type=getattr(member_type, "value", member_type),
            auth=auth,
        )
        return response["Response"]

    async def kick_member(
        self, group_id: int, membership_id: int, membership_type: Union[BungieMembershipType, int], auth: AuthData
    ) -> GroupMemberLeaveResult:
        """
        Kick a member from the given group, forcing them to reapply if they wish to re-join the group. You must have suitable permissions in the group to perform this operation.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            group_id: Group ID to kick the user from.
            membership_id: Membership ID to kick.
            membership_type: Membership type of the provided membership ID.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.kick_member(
            group_id=group_id,
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
        )
        return await GroupMemberLeaveResult.from_dict(
            data=response,
            client=self._client,
            group_id=group_id,
            membership_id=membership_id,
            membership_type=membership_type,
            auth=auth,
        )

    async def ban_member(
        self,
        data: GroupBanRequest,
        group_id: int,
        membership_id: int,
        membership_type: Union[BungieMembershipType, int],
        auth: AuthData,
    ) -> int:
        """
        Bans the requested member from the requested group for the specified period of time.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: Group ID that has the member to ban.
            membership_id: Membership ID of the member to ban from the group.
            membership_type: Membership type of the provided membership ID.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.ban_member(
            group_id=group_id,
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
            **data.to_dict()
        )
        return response["Response"]

    async def unban_member(
        self, group_id: int, membership_id: int, membership_type: Union[BungieMembershipType, int], auth: AuthData
    ) -> int:
        """
        Unbans the requested member, allowing them to re-apply for membership.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            group_id:
            membership_id: Membership ID of the member to unban from the group
            membership_type: Membership type of the provided membership ID.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.unban_member(
            group_id=group_id,
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
        )
        return response["Response"]

    async def get_banned_members_of_group(
        self, currentpage: int, group_id: int, auth: AuthData
    ) -> SearchResultOfGroupBan:
        """
        Get the list of banned members in a given group. Only accessible to group Admins and above. Not applicable to all groups. Check group features.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            currentpage: Page number (starting with 1). Each page has a fixed size of 50 entries.
            group_id: Group ID whose banned members you are fetching
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_banned_members_of_group(
            currentpage=currentpage, group_id=group_id, auth=auth
        )
        return await SearchResultOfGroupBan.from_dict(
            data=response, client=self._client, currentpage=currentpage, group_id=group_id, auth=auth
        )

    async def abdicate_foundership(
        self,
        founder_id_new: int,
        group_id: int,
        membership_type: Union[BungieMembershipType, int],
        auth: Optional[AuthData] = None,
    ) -> bool:
        """
        An administrative method to allow the founder of a group or clan to give up their position to another admin permanently.

        Args:
            founder_id_new: The new founder for this group. Must already be a group admin.
            group_id: The target group id.
            membership_type: Membership type of the provided founderIdNew.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.abdicate_foundership(
            founder_id_new=founder_id_new,
            group_id=group_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
        )
        return response["Response"]

    async def get_pending_memberships(
        self, currentpage: int, group_id: int, auth: AuthData
    ) -> SearchResultOfGroupMemberApplication:
        """
        Get the list of users who are awaiting a decision on their application to join a given group. Modified to include application info.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            currentpage: Page number (starting with 1). Each page has a fixed size of 50 items per page.
            group_id: ID of the group.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_pending_memberships(
            currentpage=currentpage, group_id=group_id, auth=auth
        )
        return await SearchResultOfGroupMemberApplication.from_dict(
            data=response, client=self._client, currentpage=currentpage, group_id=group_id, auth=auth
        )

    async def get_invited_individuals(
        self, currentpage: int, group_id: int, auth: AuthData
    ) -> SearchResultOfGroupMemberApplication:
        """
        Get the list of users who have been invited into the group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            currentpage: Page number (starting with 1). Each page has a fixed size of 50 items per page.
            group_id: ID of the group.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_invited_individuals(
            currentpage=currentpage, group_id=group_id, auth=auth
        )
        return await SearchResultOfGroupMemberApplication.from_dict(
            data=response, client=self._client, currentpage=currentpage, group_id=group_id, auth=auth
        )

    async def approve_all_pending(
        self, data: GroupApplicationRequest, group_id: int, auth: AuthData
    ) -> list[EntityActionResult]:
        """
        Approve all of the pending users for the given group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: ID of the group.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.approve_all_pending(group_id=group_id, auth=auth, **data.to_dict())
        return [
            await EntityActionResult.from_dict(data=value, client=self._client, group_id=group_id, auth=auth)
            for value in response["Response"]
        ]

    async def deny_all_pending(
        self, data: GroupApplicationRequest, group_id: int, auth: AuthData
    ) -> list[EntityActionResult]:
        """
        Deny all of the pending users for the given group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: ID of the group.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.deny_all_pending(group_id=group_id, auth=auth, **data.to_dict())
        return [
            await EntityActionResult.from_dict(data=value, client=self._client, group_id=group_id, auth=auth)
            for value in response["Response"]
        ]

    async def approve_pending_for_list(
        self, data: GroupApplicationListRequest, group_id: int, auth: AuthData
    ) -> list[EntityActionResult]:
        """
        Approve all of the pending users for the given group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: ID of the group.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.approve_pending_for_list(group_id=group_id, auth=auth, **data.to_dict())
        return [
            await EntityActionResult.from_dict(data=value, client=self._client, group_id=group_id, auth=auth)
            for value in response["Response"]
        ]

    async def approve_pending(
        self,
        data: GroupApplicationRequest,
        group_id: int,
        membership_id: int,
        membership_type: Union[BungieMembershipType, int],
        auth: AuthData,
    ) -> bool:
        """
        Approve the given membershipId to join the group/clan as long as they have applied.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: ID of the group.
            membership_id: The membership id being approved.
            membership_type: Membership type of the supplied membership ID.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.approve_pending(
            group_id=group_id,
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
            **data.to_dict()
        )
        return response["Response"]

    async def deny_pending_for_list(
        self, data: GroupApplicationListRequest, group_id: int, auth: AuthData
    ) -> list[EntityActionResult]:
        """
        Deny all of the pending users for the given group that match the passed-in .

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: ID of the group.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.deny_pending_for_list(group_id=group_id, auth=auth, **data.to_dict())
        return [
            await EntityActionResult.from_dict(data=value, client=self._client, group_id=group_id, auth=auth)
            for value in response["Response"]
        ]

    async def get_groups_for_member(
        self,
        filter: Union[GroupsForMemberFilter, int],
        group_type: Union[GroupType, int],
        membership_id: int,
        membership_type: Union[BungieMembershipType, int],
        auth: Optional[AuthData] = None,
    ) -> GetGroupsForMemberResponse:
        """
        Get information about the groups that a given member has joined.

        Args:
            filter: Filter apply to list of joined groups.
            group_type: Type of group the supplied member founded.
            membership_id: Membership ID to for which to find founded groups.
            membership_type: Membership type of the supplied membership ID.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_groups_for_member(
            filter=getattr(filter, "value", filter),
            group_type=getattr(group_type, "value", group_type),
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
        )
        return await GetGroupsForMemberResponse.from_dict(
            data=response,
            client=self._client,
            filter=filter,
            group_type=group_type,
            membership_id=membership_id,
            membership_type=membership_type,
            auth=auth,
        )

    async def recover_group_for_founder(
        self,
        group_type: Union[GroupType, int],
        membership_id: int,
        membership_type: Union[BungieMembershipType, int],
        auth: Optional[AuthData] = None,
    ) -> GroupMembershipSearchResponse:
        """
        Allows a founder to manually recover a group they can see in game but not on bungie.net

        Args:
            group_type: Type of group the supplied member founded.
            membership_id: Membership ID to for which to find founded groups.
            membership_type: Membership type of the supplied membership ID.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.recover_group_for_founder(
            group_type=getattr(group_type, "value", group_type),
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
        )
        return await GroupMembershipSearchResponse.from_dict(
            data=response,
            client=self._client,
            group_type=group_type,
            membership_id=membership_id,
            membership_type=membership_type,
            auth=auth,
        )

    async def get_potential_groups_for_member(
        self,
        filter: Union[GroupPotentialMemberStatus, int],
        group_type: Union[GroupType, int],
        membership_id: int,
        membership_type: Union[BungieMembershipType, int],
        auth: Optional[AuthData] = None,
    ) -> GroupPotentialMembershipSearchResponse:
        """
        Get information about the groups that a given member has applied to or been invited to.

        Args:
            filter: Filter apply to list of potential joined groups.
            group_type: Type of group the supplied member applied.
            membership_id: Membership ID to for which to find applied groups.
            membership_type: Membership type of the supplied membership ID.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.get_potential_groups_for_member(
            filter=getattr(filter, "value", filter),
            group_type=getattr(group_type, "value", group_type),
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
        )
        return await GroupPotentialMembershipSearchResponse.from_dict(
            data=response,
            client=self._client,
            filter=filter,
            group_type=group_type,
            membership_id=membership_id,
            membership_type=membership_type,
            auth=auth,
        )

    async def individual_group_invite(
        self,
        data: GroupApplicationRequest,
        group_id: int,
        membership_id: int,
        membership_type: Union[BungieMembershipType, int],
        auth: AuthData,
    ) -> GroupApplicationResponse:
        """
        Invite a user to join this group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: ID of the group you would like to join.
            membership_id: Membership id of the account being invited.
            membership_type: MembershipType of the account being invited.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.individual_group_invite(
            group_id=group_id,
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
            **data.to_dict()
        )
        return await GroupApplicationResponse.from_dict(
            data=response,
            client=self._client,
            group_id=group_id,
            membership_id=membership_id,
            membership_type=membership_type,
            auth=auth,
        )

    async def individual_group_invite_cancel(
        self, group_id: int, membership_id: int, membership_type: Union[BungieMembershipType, int], auth: AuthData
    ) -> GroupApplicationResponse:
        """
        Cancels a pending invitation to join a group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            group_id: ID of the group you would like to join.
            membership_id: Membership id of the account being cancelled.
            membership_type: MembershipType of the account being cancelled.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        response = await self._client.http.individual_group_invite_cancel(
            group_id=group_id,
            membership_id=membership_id,
            membership_type=getattr(membership_type, "value", membership_type),
            auth=auth,
        )
        return await GroupApplicationResponse.from_dict(
            data=response,
            client=self._client,
            group_id=group_id,
            membership_id=membership_id,
            membership_type=membership_type,
            auth=auth,
        )



# DO NOT CHANGE ANY CODE BELOW
# This file is generated automatically by `generate_api_schema.py` and will be overwritten
# Instead, change functions / models by subclassing them in the `./overwrites/` folder. They will be used instead.

from typing import TYPE_CHECKING, Optional

from bungio.models.base import BaseModel, custom_define, custom_field

if TYPE_CHECKING:
    from bungio.models import DestinyInventoryItemDefinition, DestinyObjectiveProgress


@custom_define()
class DestinyItemPlugBase(BaseModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        can_insert: If true, this plug has met all of its insertion requirements. Big if true.
        enable_fail_indexes: If a plug is not enabled, this will be populated with indexes into the plug item definition's plug.enabledRules property, so that you can show the reasons why it is not enabled. This list will be empty if the plug is enabled.
        enabled: If true, this plug will provide its benefits while inserted.
        insert_fail_indexes: If the plug cannot be inserted for some reason, this will have the indexes into the plug item definition's plug.insertionRules property, so you can show the reasons why it can't be inserted. This list will be empty if the plug can be inserted.
        plug_item_hash: The hash identifier of the DestinyInventoryItemDefinition that represents this plug.
        manifest_plug_item_hash: Manifest information for `plug_item_hash`
    """

    can_insert: bool = custom_field()
    enable_fail_indexes: list[int] = custom_field(metadata={"type": """list[int]"""})
    enabled: bool = custom_field()
    insert_fail_indexes: list[int] = custom_field(metadata={"type": """list[int]"""})
    plug_item_hash: int = custom_field()
    manifest_plug_item_hash: Optional["DestinyInventoryItemDefinition"] = custom_field(default=None)


@custom_define()
class DestinyItemPlug(BaseModel):
    """
    _No description given by bungie._

    Tip: Manifest Information
        This model has some attributes which can be filled with additional information found in the manifest (`manifest_...`).
        Without additional work, these attributes will be `None`, since they require additional requests and database lookups.

        To fill the manifest dependent attributes, either:

        - Run `await ThisClass.fetch_manifest_information()`, see [here](/API Reference/Models/base)
        - Set `Client.always_return_manifest_information` to `True`, see [here](/API Reference/client)

    Attributes:
        can_insert: If true, this plug has met all of its insertion requirements. Big if true.
        enable_fail_indexes: If a plug is not enabled, this will be populated with indexes into the plug item definition's plug.enabledRules property, so that you can show the reasons why it is not enabled. This list will be empty if the plug is enabled.
        enabled: If true, this plug will provide its benefits while inserted.
        insert_fail_indexes: If the plug cannot be inserted for some reason, this will have the indexes into the plug item definition's plug.insertionRules property, so you can show the reasons why it can't be inserted. This list will be empty if the plug can be inserted.
        plug_item_hash: The hash identifier of the DestinyInventoryItemDefinition that represents this plug.
        plug_objectives: Sometimes, Plugs may have objectives: these are often used for flavor and display purposes, but they can be used for any arbitrary purpose (both fortunately and unfortunately). Recently (with Season 2) they were expanded in use to be used as the "gating" for whether the plug can be inserted at all. For instance, a Plug might be tracking the number of PVP kills you have made. It will use the parent item's data about that tracking status to determine what to show, and will generally show it using the DestinyObjectiveDefinition's progressDescription property. Refer to the plug's itemHash and objective property for more information if you would like to display even more data.
        manifest_plug_item_hash: Manifest information for `plug_item_hash`
    """

    can_insert: bool = custom_field()
    enable_fail_indexes: list[int] = custom_field(metadata={"type": """list[int]"""})
    enabled: bool = custom_field()
    insert_fail_indexes: list[int] = custom_field(metadata={"type": """list[int]"""})
    plug_item_hash: int = custom_field()
    plug_objectives: list["DestinyObjectiveProgress"] = custom_field(
        metadata={"type": """list[DestinyObjectiveProgress]"""}
    )
    manifest_plug_item_hash: Optional["DestinyInventoryItemDefinition"] = custom_field(default=None)



from datetime import datetime
from typing import TYPE_CHECKING, AsyncGenerator, Optional, Union

from bungio.models.base import ClientMixin, FuzzyAttrFinder, custom_define

if TYPE_CHECKING:
    # AUTOMATIC IMPORTS START
    from bungio.models import (
        BungieMembershipType,
        BungieRewardDisplay,
        DestinyComponentType,
        DestinyHistoricalStatsAccountResult,
        DestinyItemResponse,
        DestinyLeaderboard,
        DestinyLinkedProfilesResponse,
        DestinyProfileResponse,
        DestinyStatsGroupType,
        GetGroupsForMemberResponse,
        GroupApplicationRequest,
        GroupApplicationResponse,
        GroupBanRequest,
        GroupMemberLeaveResult,
        GroupMembershipSearchResponse,
        GroupPotentialMembershipSearchResponse,
        GroupPotentialMemberStatus,
        GroupsForMemberFilter,
        GroupType,
        RuntimeGroupMemberType,
        UserMembershipData,
    )

    # AUTOMATIC IMPORTS END
    from bungio.models.auth import AuthData
    from bungio.models.basic import DestinyCharacter
    from bungio.models.bungie import DestinyActivityModeType
    from bungio.models.overwrites import DestinyHistoricalStatsPeriodGroup


__all__ = ("DestinyUserMixin",)


@custom_define()
class DestinyUserMixin(ClientMixin, FuzzyAttrFinder):
    @property
    def full_bungie_name(self) -> str:
        """
        Return the formatted bungie name like it is seen in-game. This includes the four numbers

        Returns:
            The full bungie name
        """

        return f"""{self.bungie_global_display_name}#{str(self.bungie_global_display_name_code).zfill(4)}"""

    async def yield_activity_history(
        self,
        mode: Union["DestinyActivityModeType", int],
        earliest_allowed_datetime: Optional[datetime] = None,
        latest_allowed_datetime: Optional[datetime] = None,
        auth: Optional["AuthData"] = None,
    ) -> AsyncGenerator["DestinyHistoricalStatsPeriodGroup", None]:
        """
        Yields account activity history, no matter the character. Sorted by date descending, the latest one first.

        Args:
            mode: A filter for the activity mode to be returned. None returns all activities. See the documentation for DestinyActivityModeType for valid values, and pass in string representation.
            earliest_allowed_datetime: The earliest time the activity is allowed to have, fe. only entries after the 1/1/2020.
            latest_allowed_datetime: The latest time the activity is allowed to have, fe. only entries before the 1/1/2020.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            A generator for the model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        # get character ids and gen functions
        # use the stats page to also get deleted chars
        data = await self.get_historical_stats_for_account(groups=[DestinyStatsGroupType.NONE], auth=auth)

        characters = [
            DestinyCharacter(
                membership_id=self._fuzzy_getattr("membership_id"),
                membership_type=self._fuzzy_getattr("membership_type"),
                character_id=character.character_id,
            )
            for character in data.characters
        ]

        funcs = {
            i: character.yield_activity_history(
                mode=mode,
                earliest_allowed_datetime=earliest_allowed_datetime,
                latest_allowed_datetime=latest_allowed_datetime,
                auth=auth,
            )
            for i, character in enumerate(characters)
        }

        # gen the first values
        entries = {i: await anext(func, None) for i, func in funcs.items()}

        # loop through all generators and return the newest values
        while True:
            # calculate the newest entry
            newest: Optional["DestinyHistoricalStatsPeriodGroup"] = None
            newest_index: int = 0
            for i, entry in entries.items():
                if entry is not None and (newest is None or entry.period > newest.period):
                    newest = entry
                    newest_index = i

            if newest:
                yield newest

                # get a new value from the func that just yielded
                entries[newest_index] = await anext(funcs[newest_index], None)
            else:
                break

    # DO NOT CHANGE ANY CODE BELOW. Automatically generated and overwritten

    async def get_membership_data_by_id(self, auth: Optional["AuthData"] = None) -> "UserMembershipData":
        """
        Returns a list of accounts associated with the supplied membership ID and membership type. This will include all linked accounts (even when hidden) if supplied credentials permit it.

        Args:
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.get_membership_data_by_id(
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def edit_group_membership(
        self, group_id: int, member_type: Union["RuntimeGroupMemberType", int], auth: "AuthData"
    ) -> int:
        """
        Edit the membership type of a given member. You must have suitable permissions in the group to perform this operation.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            group_id: ID of the group to which the member belongs.
            member_type: New membertype for the specified member.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.edit_group_membership(
            group_id=group_id,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            member_type=member_type,
            auth=auth,
        )

    async def kick_member(self, group_id: int, auth: "AuthData") -> "GroupMemberLeaveResult":
        """
        Kick a member from the given group, forcing them to reapply if they wish to re-join the group. You must have suitable permissions in the group to perform this operation.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            group_id: Group ID to kick the user from.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.kick_member(
            group_id=group_id,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def ban_member(self, data: "GroupBanRequest", group_id: int, auth: "AuthData") -> int:
        """
        Bans the requested member from the requested group for the specified period of time.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: Group ID that has the member to ban.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.ban_member(
            data=data,
            group_id=group_id,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def unban_member(self, group_id: int, auth: "AuthData") -> int:
        """
        Unbans the requested member, allowing them to re-apply for membership.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            group_id:
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.unban_member(
            group_id=group_id,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def approve_pending(self, data: "GroupApplicationRequest", group_id: int, auth: "AuthData") -> bool:
        """
        Approve the given membershipId to join the group/clan as long as they have applied.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: ID of the group.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.approve_pending(
            data=data,
            group_id=group_id,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def get_groups_for_member(
        self,
        filter: Union["GroupsForMemberFilter", int],
        group_type: Union["GroupType", int],
        auth: Optional["AuthData"] = None,
    ) -> "GetGroupsForMemberResponse":
        """
        Get information about the groups that a given member has joined.

        Args:
            filter: Filter apply to list of joined groups.
            group_type: Type of group the supplied member founded.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.get_groups_for_member(
            filter=filter,
            group_type=group_type,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def recover_group_for_founder(
        self, group_type: Union["GroupType", int], auth: Optional["AuthData"] = None
    ) -> "GroupMembershipSearchResponse":
        """
        Allows a founder to manually recover a group they can see in game but not on bungie.net

        Args:
            group_type: Type of group the supplied member founded.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.recover_group_for_founder(
            group_type=group_type,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def get_potential_groups_for_member(
        self,
        filter: Union["GroupPotentialMemberStatus", int],
        group_type: Union["GroupType", int],
        auth: Optional["AuthData"] = None,
    ) -> "GroupPotentialMembershipSearchResponse":
        """
        Get information about the groups that a given member has applied to or been invited to.

        Args:
            filter: Filter apply to list of potential joined groups.
            group_type: Type of group the supplied member applied.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.get_potential_groups_for_member(
            filter=filter,
            group_type=group_type,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def individual_group_invite(
        self, data: "GroupApplicationRequest", group_id: int, auth: "AuthData"
    ) -> "GroupApplicationResponse":
        """
        Invite a user to join this group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            data: The required data for this request.
            group_id: ID of the group you would like to join.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.individual_group_invite(
            data=data,
            group_id=group_id,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def individual_group_invite_cancel(self, group_id: int, auth: "AuthData") -> "GroupApplicationResponse":
        """
        Cancels a pending invitation to join a group.

        Warning: Requires Authentication.
            Required oauth2 scopes: AdminGroups

        Args:
            group_id: ID of the group you would like to join.
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.individual_group_invite_cancel(
            group_id=group_id,
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def get_bungie_rewards_for_platform_user(self, auth: "AuthData") -> dict[str, "BungieRewardDisplay"]:
        """
        Returns the bungie rewards for the targeted user when a platform membership Id and Type are used.

        Warning: Requires Authentication.
            Required oauth2 scopes: ReadAndApplyTokens

        Args:
            auth: Authentication information.

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.get_bungie_rewards_for_platform_user(
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            auth=auth,
        )

    async def get_linked_profiles(
        self, get_all_memberships: bool, auth: Optional["AuthData"] = None
    ) -> "DestinyLinkedProfilesResponse":
        """
        Returns a summary information about all profiles linked to the requesting membership type/membership ID that have valid Destiny information. The passed-in Membership Type/Membership ID may be a Bungie.Net membership or a Destiny membership. It only returns the minimal amount of data to begin making more substantive requests, but will hopefully serve as a useful alternative to UserServices for people who just care about Destiny data. Note that it will only return linked accounts whose linkages you are allowed to view.

        Args:
            get_all_memberships: (optional) if set to 'true', all memberships regardless of whether they're obscured by overrides will be returned. Normal privacy restrictions on account linking will still apply no matter what.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.get_linked_profiles(
            membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            get_all_memberships=get_all_memberships,
            auth=auth,
        )

    async def get_profile(
        self, components: list[Union["DestinyComponentType", int]], auth: Optional["AuthData"] = None
    ) -> "DestinyProfileResponse":
        """
        Returns Destiny Profile information for the supplied membership.

        Args:
            components: A comma separated list of components to return (as strings or numeric values). See the DestinyComponentType enum for valid components to request. You must request at least one component to receive results.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.get_profile(
            destiny_membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            components=components,
            auth=auth,
        )

    async def get_item(
        self,
        item_instance_id: int,
        components: list[Union["DestinyComponentType", int]],
        auth: Optional["AuthData"] = None,
    ) -> "DestinyItemResponse":
        """
        Retrieve the details of an instanced Destiny Item. An instanced Destiny item is one with an ItemInstanceId. Non-instanced items, such as materials, have no useful instance-specific details and thus are not queryable here.

        Args:
            item_instance_id: The Instance ID of the destiny item.
            components: A comma separated list of components to return (as strings or numeric values). See the DestinyComponentType enum for valid components to request. You must request at least one component to receive results.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.get_item(
            destiny_membership_id=self._fuzzy_getattr("membership_id"),
            item_instance_id=item_instance_id,
            membership_type=self._fuzzy_getattr("membership_type"),
            components=components,
            auth=auth,
        )

    async def get_leaderboards(
        self, maxtop: int, modes: str, statid: str, auth: Optional["AuthData"] = None
    ) -> dict[str, dict[str, "DestinyLeaderboard"]]:
        """
        Gets leaderboards with the signed in user's friends and the supplied destinyMembershipId as the focus. PREVIEW: This endpoint has not yet been implemented. It is being returned for a preview of future functionality, and for public comment/suggestion/preparation.

        Args:
            maxtop: Maximum number of top players to return. Use a large number to get entire leaderboard.
            modes: List of game modes for which to get leaderboards. See the documentation for DestinyActivityModeType for valid values, and pass in string representation, comma delimited.
            statid: ID of stat to return rather than returning all Leaderboard stats.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.get_leaderboards(
            destiny_membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            maxtop=maxtop,
            modes=modes,
            statid=statid,
            auth=auth,
        )

    async def get_historical_stats_for_account(
        self, groups: list[Union["DestinyStatsGroupType", int]], auth: Optional["AuthData"] = None
    ) -> "DestinyHistoricalStatsAccountResult":
        """
        Gets aggregate historical stats organized around each character for a given account.

        Args:
            groups: Groups of stats to include, otherwise only general stats are returned. Comma separated list is allowed. Values: General, Weapons, Medals.
            auth: Authentication information. Required when users with a private profile are queried, or when Bungie feels like it

        Returns:
            The model which is returned by bungie. [General endpoint information.](https://bungie-net.github.io/multi/index.html)
        """

        return await self._client.api.get_historical_stats_for_account(
            destiny_membership_id=self._fuzzy_getattr("membership_id"),
            membership_type=self._fuzzy_getattr("membership_type"),
            groups=groups,
            auth=auth,
        )
